# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# **************************************************************************

import bpy
import bmesh

import numpy as np
from collections import deque
from contextlib import contextmanager

import blendgamer.pygamer as pygamer
import blendgamer.pygamer.surfacemesh as sm

# DEFINITIONS
UNSETID = 0  # ID value for unset markers
# Default should match MeshPolygonIntProperty default
UNSETMARKER = -1  # Marker value to output for unset faces
materialNamer = lambda bnd_id: "%s_mat" % (bnd_id)
##


class ObjectMode:
    """Class defining a Blender Object mode context for with statement semantics

    Attributes
    ----------
    mode : str
        Enum of Blender's mode
    """

    def __enter__(self):
        """
        Enter runtime context to cache current mode and switch to Object mode.
        """
        self.mode = bpy.context.active_object.mode
        bpy.ops.object.mode_set(mode="OBJECT")

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit runtime context and restore mode prior to enter.
        """
        # print("Changing to %s mode"%(self.mode))
        bpy.ops.object.mode_set(mode=self.mode)
        # Uncomment to handle exceptions
        # return True


@contextmanager
def BMeshContext(obj):
    if obj.type != "MESH":
        raise RuntimeError("Expected an object of MESH type.")

    me = obj.data
    if obj.mode == "EDIT":
        bm = bmesh.from_edit_mesh(me)
    else:
        bm = bmesh.new()
        bm.from_mesh(me)

    yield bm

    if obj.mode == "EDIT":
        if bpy.app.version < (2, 80, 0):
            bmesh.update_edit_mesh(me, tessface=True)
        else:
            bmesh.update_edit_mesh(me, loop_triangles=True)
        # BMesh from edit meshes should never be freed!
        # https://developer.blender.org/T39121
    else:
        bm.to_mesh(me)
        bm.free()
        del bm


@contextmanager
def copiedBMeshContext(obj, transform=False, triangulate=False, apply_modifiers=False):
    """Context for a bmesh copied from the mesh. Changes are not written back to the original mesh."""
    if obj.type != "MESH":
        raise RuntimeError("Expected object of MESH type.")

    if apply_modifiers and obj.modifiers:
        if bpy.app.version < (2, 80, 0):
            me = obj.to_mesh(
                bpy.context.scene, apply_modifiers=True, settings="PREVIEW"
            )
        else:
            me = obj.to_mesh()
        bm = bmesh.new()
        bm.from_mesh(me)
        bpy.data.meshes.remove(me)
    else:
        me = obj.data
        if obj.mode == "EDIT":
            bm_orig = bmesh.from_edit_mesh(me)
            bm = bm_orig.copy()
        else:
            bm = bmesh.new()
            bm.from_mesh(me)

    if transform:
        bm.transform(obj.matrix_world)

    if triangulate:
        bmesh.ops.triangulate(bm, faces=bm.faces)

    yield bm

    bm.free()
    del bm


def get_material_by_bnd_id(boundary_id):
    """
    Gets a boundary material from bpy.data.materials by boundary ID.

    Parameters
    ----------
    boundary_id : int
        The boundary ID to search for.

    Returns
    -------
    bpy.types.Material or None
        Returns the material if found or None if not.
    """
    mats = bpy.data.materials
    for mat in mats:
        if mat.gamer.boundary_id == boundary_id:
            return mat
    return None


def getBndUnsetMat():
    """
    Get the material for unset boundaries.

    Initializes the material if it does not already exist.

    Returns
    -------
    bpy.types.Material
        Returns the material corresponding to unmarked boundaries.
    """
    bnd_unset_mat = get_material_by_bnd_id(UNSETID)
    # if bnd_unset_mat is not defined, then create it
    if bnd_unset_mat is None:
        bnd_unset_mat = bpy.data.materials.new("bnd_unset_mat")
        bnd_unset_mat.use_fake_user = True
        bnd_unset_mat.gamer.boundary_id = UNSETID

    return bnd_unset_mat


def getMarkerLayer(obj):
    """Gets the boundary marker layer data of an object

    Parameters
    ----------
    obj : bpy.types.Object
        Object of interest

    Returns
    -------
    bpy.types.bpy_prop_collection
        Containing `bpy.types.MeshPolygonIntProperty` values corresponding to marker boundaryIDs

    Raises
    ------
    RuntimeError
        Data layers can only be accessed in 'OBJECT' mode
    """
    if obj.mode != "OBJECT":
        raise RuntimeError(
            "Blender Layers (Markers) can only be accessed in 'OBJECT' mode."
        )
    markerLayer = obj.data.polygon_layers_int.get("marker")
    # If the layer doesn't exist yet, create it!
    if not markerLayer:
        markerLayer = obj.data.polygon_layers_int.new(name="marker")
    return markerLayer.data


def getBMeshMarkerLayer(bm):
    """Get the boundary marker layer from a BMesh

    Args:
        bm (bpy.BMesh): BMesh object

    Returns:
        MarkerLayer: Marker layer
    """
    markerLayer = bm.faces.layers.int.get("marker")
    if not markerLayer:
        markerLayer = bm.faces.layers.int.new("marker")
    return markerLayer


def getCurvatureLayer(obj, algo, curvatureType):
    """Get the data layer corresponding to a specific curvature type

    Parameters
    ----------
    obj : bpy.types.Object
        Object of interest
    algo : TYPE
        Description
    curvatureType : TYPE
        Description

    Returns
    -------
    TYPE
        Description

    Raises
    ------
    RuntimeError
    Description
    """
    name = "%s%s" % (algo, curvatureType)
    if obj.mode != "OBJECT":
        raise RuntimeError(
            "Blender Layers (%s) can only be accessed in 'OBJECT' mode." % (name)
        )
    layer = obj.data.vertex_layers_float.get(name)
    # If the layer doesn't exist yet, create it!
    if not layer:
        layer = obj.data.vertex_layers_float.new(name=name)
    return layer.data


def getActiveMeshObject():
    """
    Returns the active object if it is a mesh or None

    Returns
    -------
    bpy.types.Object
        Mesh object of interest

    Raises
    ------
    RuntimeError
        If active object is not a Mesh or no active object exists.
    """
    obj = bpy.context.active_object
    if obj:
        if obj.type == "MESH":
            #  # Auto select object if in EDIT mode
            # # This prevents being 'selected' in EDIT mode but really unselected
            # if obj.mode == 'EDIT':
            #     obj.select = True
            # if obj.select:
            return obj
        else:
            raise RuntimeError(
                "Active object is not a MESH. Please select a MESH to use this feature."
            )
    else:
        raise RuntimeError(
            "No active object! Please select a MESH to use this feature."
        )


def getMeshVertices(obj, get_selected_vertices=False):
    """
    Get the vertices of mesh object

    Parameters
    ----------
    obj : TYPE
        Description
    get_selected_vertices : bool, optional
        Description

    Returns
    -------
    TYPE
        Description
    """
    mesh = obj.data
    vertToVec = lambda v: [v[0], v[1], v[2]]

    vertices = [vertToVec(v.co) for v in mesh.vertices]

    if get_selected_vertices:
        selected_indices = [v.index for v in mesh.vertices if v.select and not v.hide]
        return vertices, selected_indices
    else:
        return vertices


def blender_to_gamer(obj=None, map_boundaries=False, autocorrect_normals=True):
    """Construct a GAMer mesh from a triangulated blender mesh

    Args:
        obj (, optional): Object to convert. Defaults to None which gets the active mesh object.
        map_boundaries (bool, optional): True if boundary values should be mapped to markers instead of boundary_id. Defaults to False.
        autocorrect_normals (bool, optional): Automatically flip normals so that volume is positive. Defaults to True.

    Raises:
        RuntimeError: Complains if something prevents conversion to a valid GAMer mesh

    Returns:
        gamer.SurfaceMesh: GAMer surface mesh from Blender mesh
    """
    # Get the selected mesh
    if not obj:
        obj = getActiveMeshObject()
        # Ensure object is good
        if not obj:
            return False

    with ObjectMode():
        # Grab vertices
        vertices, selected_vertices = getMeshVertices(obj, get_selected_vertices=True)
        # Get world location and offset each vertex with this value
        gmesh = sm.SurfaceMesh()  # Init GAMer SurfaceMesh

        def addVertex(co, sel):  # functor to addVertices
            gmesh.addVertex(
                sm.Vertex(
                    co[0],  # x position
                    co[1],  # y position
                    co[2],  # z position
                    0,
                    bool(sel),  # selected flag
                )
            )

        # If all vertices are selected
        if len(selected_vertices) == len(vertices):
            selected_vertices = np.ones(len(vertices), dtype=bool)
        else:
            selection = np.zeros(len(vertices), dtype=bool)
            selection[selected_vertices] = 1  # Set selected to True
            selected_vertices = selection

        # Zip args and pass to addVertex functor
        [addVertex(*args) for args in zip(vertices, selected_vertices)]

        ml = getMarkerLayer(obj)
        # Transfer boundary information
        if map_boundaries:
            bdryMap = {UNSETID: UNSETMARKER}
            for bdry in obj.gamer.markers.boundary_list:
                bdryMap[bdry.boundary_id] = bdry.marker
            boundaries = [bdryMap[item.value] for item in ml.values()]
        else:
            boundaries = [item.value for item in ml.values()]

        edges = obj.data.edges
        for edge in edges:
            gmesh.insertEdge(list(edge.vertices), sm.Edge(bool(edge.select)))

        # Get list of faces
        faces = obj.data.polygons

        # Transfer data from blender mesh to gamer mesh
        for face in faces:
            vertices = deque(face.vertices)
            if len(vertices) != 3:
                raise RuntimeError(
                    "Encountered a non-triangular face. GAMer only works with triangulated meshes."
                )

            # Get the orientation from Blender
            max_val = max(vertices)
            max_idx = vertices.index(max_val)
            if max_idx != 2:
                vertices.rotate(2 - max_idx)
            orientation = -1  # 1 < 0 < 2
            if vertices[0] < vertices[1]:
                # 0 < 1 < 2
                orientation = 1
            gmesh.insertFace(
                list(vertices),
                sm.Face(orientation, boundaries[face.index], bool(face.select)),
            )
    # Ensure all face orientations are set
    gmesh.init_orientation()
    gmesh.check_orientation()
    vol = gmesh.getVolume()
    if vol < 0:
        if autocorrect_normals:
            gmesh.flipNormals()
        else:
            raise RuntimeError(
                "Mesh has negative volume. Recompute normals to be outward facing."
            )
    return gmesh


def gamer_to_blender(gmesh, obj=None, mesh_name="gamer_improved"):
    """Pass GAMer Surface Mesh back to Blender]

    Args:
        gmesh (gamer.SurfaceMesh): GAMer surface mesh
        obj (bpy.types.Object, optional):  Object to operate on. Active object is used if unset. Defaults to None.
        mesh_name (str, optional): Name of the new mesh data block. Defaults to "gamer_improved".

    Raises:
        RuntimeError: If an undefined behavior encountered.
    """

    # Check arguments
    if not isinstance(gmesh, sm.SurfaceMesh):
        raise RuntimeError(
            "gamer_to_blender expected a pygamer.surfacemesh.SurfaceMesh object."
        )

    if not obj:
        obj = getActiveMeshObject()
        if not obj:
            raise RuntimeError("Active object is not a MESH.")

    mode = obj.mode

    verts = []  # Array of vertex coordinates
    idxMap = {}  # Dictionary of gamer indices to renumbered indices
    for i, vid in enumerate(gmesh.vertexIDs):
        v = vid.data()
        verts.append((v[0], v[1], v[2]))
        idxMap[gmesh.getName(vid)[0]] = i  # Reindex starting at 0

    faces = []
    selectedFaces = []
    markersList = []
    for i, fid in enumerate(gmesh.faceIDs):
        fName = gmesh.getName(fid)
        face = fid.data()

        if face.selected:
            selectedFaces.append(i)

        if face.orientation == 1:
            faces.append((idxMap[fName[0]], idxMap[fName[1]], idxMap[fName[2]]))
        else:
            faces.append((idxMap[fName[2]], idxMap[fName[1]], idxMap[fName[0]]))
        markersList.append(face.marker)

    with ObjectMode():
        if bpy.app.version < (2, 81, 0):
            # Save old mesh data...
            # Can't delete yet or object goes out of context
            oldmesh = obj.data
            # Create new mesh and delete old
            mesh = bpy.data.meshes.new(mesh_name)
            mesh.from_pydata(verts, [], faces)
            # mesh.validate(verbose=True)
            mesh.update()
            # mesh.calc_normals()
            obj.data = mesh
            bpy.data.meshes.remove(oldmesh)
        else:
            mesh = obj.data
            mesh.clear_geometry()
            mesh.from_pydata(verts, [], faces)
            mesh.update()
            # mesh.calc_normals()

        ml = getMarkerLayer(obj)
        for i, marker in enumerate(markersList):
            ml[i].value = marker

    # Repaint boundaries
    obj.gamer.markers.repaint_boundaries(bpy.context)
    # Deselect all first
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.select_all(action="DESELECT")
    bpy.ops.object.mode_set(mode="OBJECT")

    bm = bmesh_from_object(obj)

    with BMeshContext(obj) as bm:
        bm.faces.ensure_lookup_table()
        for f in selectedFaces:
            bm.faces[f].select_set(True)

    bpy.ops.object.mode_set(mode=mode)


## Following functions are from 3D Print Addon
def clean_float(text):
    # strip trailing zeros: 0.000 -> 0.0
    index = text.rfind(".")
    if index != -1:
        index += 2
        head, tail = text[:index], text[index:]
        tail = tail.rstrip("0")
        text = head + tail
    return text


def bmesh_from_object(obj):
    """
    Object/Edit Mode get mesh, use bmesh_to_object() to write back.

    Parameters
    ----------
    obj : TYPE
        Description

    Returns
    -------
    TYPE
        Description
    """
    me = obj.data
    is_editmode = obj.mode == "EDIT"
    if is_editmode:
        bm = bmesh.from_edit_mesh(me)
    else:
        bm = bmesh.new()
        bm.from_mesh(me)
    return bm


def bmesh_to_object(obj, bm):
    """
    Object/Edit Mode update the object.
    NOTE that this free's the bmesh

    Parameters
    ----------
    obj : TYPE
        Description
    bm : TYPE
        Description
    """
    me = obj.data
    is_editmode = obj.mode == "EDIT"
    if is_editmode:
        bmesh.update_edit_mesh(me, True)
    else:
        bm.to_mesh(me)
        bm.free()
    # grr... cause an update
    if me.vertices:
        me.vertices[0].co[0] = me.vertices[0].co[0]


def make_annotations(cls):
    """Converts class fields to annotations if running with Blender 2.8

    This is a helper function allow code from 2.79 to work with 2.8x versions
    of Blender. All calls in register and unregister to pass through this.

    Returns
    -------
    class
        Converted class
    """
    if bpy.app.version < (2, 80):
        return cls

    if bpy.app.version >= (2, 93, 0):
        bl_props = {
            k: v
            for k, v in cls.__dict__.items()
            if isinstance(v, bpy.props._PropertyDeferred)
        }
    else:
        bl_props = {k: v for k, v in cls.__dict__.items() if isinstance(v, tuple)}

    if bl_props:
        if "__annotations__" not in cls.__dict__:
            setattr(cls, "__annotations__", {})
        annotations = cls.__dict__["__annotations__"]
        for k, v in bl_props.items():
            annotations[k] = v
            delattr(cls, k)
    return cls



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# **************************************************************************

import bpy
import sys
import re

from bpy.props import PointerProperty
from ast import literal_eval
from blendgamer.util import *


class GAMER_OT_prompt_update(bpy.types.Operator):
    bl_idname = "gamer.prompt_update"
    bl_label = "Warn to update GAMer addon"
    bl_options = {"BLOCKING", "INTERNAL"}

    def execute(self, context):
        self.report(
            {"WARNING"}, "Blendfile was generated with a newer version of GAMer."
        )
        return {"FINISHED"}


class GAMER_OT_prompt_old_version(bpy.types.Operator):
    bl_idname = "gamer.prompt_old_version"
    bl_label = "Warn that GAMer cannot convert file automatically"
    bl_options = {"BLOCKING", "INTERNAL"}

    def execute(self, context):
        self.report(
            {"WARNING"},
            "Blendfile was generated with a version that does not support automatic conversion.",
        )
        return {"FINISHED"}


class GAMER_OT_update_to_2_0_1_from_v_0_1(bpy.types.Operator):
    bl_idname = "gamer.update_to_2_0_1_from_v_0_1"
    bl_label = "Update from v0.1 to v2.0.1"
    bl_description = "Update GAMer version to 2.0.1 from v0.1"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        for obj in context.scene.objects:
            # bpy.context.scene.objects.active = obj
            if obj.type == "MESH":
                print("\n" + "=" * 30 + "\n", obj, "\n" + "=" * 30)
                if not "boundaries" in obj:
                    continue
                hidden = False
                if obj.hide:
                    hidden = True
                    obj.hide = False
                obj.gamer.markers.remove_all_boundaries(context)
                for key, bdry in obj["boundaries"].items():
                    print("Migrating boundary: %s" % (key))

                    # First move the material...
                    mat = bpy.data.materials[key + "_mat"]
                    newBdryID = context.scene.gamer.boundary_id_counter + 1
                    mat.name = materialNamer(newBdryID)
                    mat.gamer.boundary_id = newBdryID
                    mat.use_fake_user = True

                    obj.gamer.markers.add_boundary(context)
                    newBdry = obj.gamer.markers.boundary_list[
                        obj.gamer.markers.active_bnd_index
                    ]

                    newBdry.boundary_name = "NewBoundaryFrom_%s" % (key)
                    newBdry.marker = bdry["marker"]

                    # # Deselect all
                    bpy.ops.object.mode_set(mode="EDIT")
                    bpy.ops.mesh.select_all(action="DESELECT")
                    bpy.ops.object.mode_set(mode="OBJECT")

                    # Select faces of interest
                    for faces in bdry["faces"].values():
                        for i in faces:
                            obj.data.polygons[i].select = True
                    bpy.ops.object.mode_set(mode="EDIT")
                    newBdry.assign_boundary_faces(context)
                    bpy.ops.object.mode_set(mode="OBJECT")
                del obj["boundaries"]
                if "id_counter" in obj.gamer:
                    del obj.gamer["id_counter"]
                if "include" in obj.gamer:
                    del obj.gamer["include"]
                obj.hide = hidden
        context.scene.gamer.gamer_version = "(2,0,1)"
        checkVersion()
        return {"FINISHED"}


def migrate2_0_1__2_0_6():
    """
    Migrate metadata formats from 2.0.1 <= v < 2.0.6 to v2.0.6
    """
    for obj in bpy.data.objects:
        if obj.type == "MESH":
            # First initialize the data-blocks
            markerProp = obj.gamer.markers
            bl = markerProp.boundary_list

            # Move boundary_list
            if "boundary_list" in obj["gamer"]:
                # Migrate boundary marker info
                obj["gamer"]["markers"]["boundary_list"] = obj["gamer"]["boundary_list"]
                del obj["gamer"]["boundary_list"]
            # Remove active_bnd_index
            if "active_bnd_index" in obj["gamer"]:
                del obj["gamer"]["active_bnd_index"]
            # Remove include
            if "include" in obj["gamer"]:
                del obj["gamer"]["include"]


def migrate2_0_6__2_0_7():
    """
    Migrate metadata formats from 2.0.6 <= v < 2.0.7 to v2.0.7
    """
    scene = bpy.context.scene
    domain_list = scene.gamer.tet_group.domain_list
    for i in range(len(domain_list) - 1, -1, -1):
        domain = domain_list[i]
        # print("Processing: ", domain["object_name"])

        if "object_name" in domain:
            if domain["object_name"] not in scene.objects: 
                print("Domain %s is missing... removing it from the domain list" % (domain["object_name"]))
                scene.gamer.tet_group.remove_domain_by_index(i)
            else:
                obj = scene.objects[domain["object_name"]]
                if obj is not None:
                    domain.object_pointer = obj
                    del domain["object_name"]
                else:
                    print("Domain %s is missing from the scene... removing it from the domain list" % (domain["object_name"]))
                    scene.gamer.tet_group.remove_domain_by_index(i)
        else:
            if domain.object_pointer is None:
                print("Unrecoverable domain found... removing it")
                scene.gamer.tet_group.remove_domain_by_index(i)

    scene.gamer.tet_group.validate_domain_objects(bpy.context, None)


def getGamerVersion():
    """
    Gets the BlendGAMer version

    Returns:
    tuple
        Tuple corresponding to the BlendGAMer version
    """
    return sys.modules["blendgamer"].bl_info.get("version", (-1, -1, -1))


def checkVersion():
    """
    Checks the blendfile metadata version with the current addon version.

    This function checks the version and attempts to migrate from older
    metadata formats when possible.

    Warnings
    --------
    This function toggles versionerror to trigger UI changes for the end user
    if it cannot migrate metadata automatically.
    """
    scene = bpy.context.scene
    print("Blendfile contains GAMer v%s metadata" % (scene.gamer.gamer_version))

    fileVer = scene.gamer.gamer_version
    isTupleStr = re.compile("\(.*\)")
    if isTupleStr.match(fileVer):
        fileVer = literal_eval(fileVer)
    else:
        fileVer = tuple(fileVer.split("."))

    currVer = getGamerVersion()
    scene.gamer.versionerror = compare_version(fileVer, currVer)

    while scene.gamer.versionerror < 0:
        if scene.gamer.versionerror == -1:
            # Throw an error for older versions of GAMer
            if compare_version(fileVer, (2, 0, 0)) < 0:
                print("trigger old version")
                bpy.ops.gamer.prompt_old_version()
                break

            # Update from 2.0.0 to 2.0.1
            elif compare_version(fileVer, (2, 0, 0)) == 0:
                newver = (2, 0, 1)
                print(
                    "Metadata version is out of date.",
                    "Migrating from v(2,0,0) to v%s" % (str(newver)),
                )
                for obj in bpy.data.objects:
                    if obj.type == "MESH":
                        # Migrate name to boundary_name
                        for bdry in obj.gamer.markers.boundary_list:
                            bdry.boundary_name = bdry.name
                            bdry.name = str(bdry.boundary_id)
                            if "boundaries" in obj.keys():
                                del obj["boundaries"]
                scene.gamer.gamer_version = str(newver)

            # Update 2.0.1--2.0.5 metadata to 2.0.6
            elif (
                compare_version(fileVer, (2, 0, 1)) >= 0
                and compare_version(fileVer, (2, 0, 6)) < 0
            ):
                newver = (2, 0, 6)
                print("Migrating from v%s to v%s" % (str(fileVer), str(newver)))
                migrate2_0_1__2_0_6()
                scene.gamer.gamer_version = str(newver)

            # Update 2.0.6 to 2.0.7
            elif (
                compare_version(fileVer, (2, 0, 6)) >= 0
                and compare_version(fileVer, (2, 0, 7)) < 0
            ):
                newver = (2, 0, 7)
                print("Migrating from v%s to v%s" % (str(fileVer), str(newver)))
                migrate2_0_6__2_0_7()
                scene.gamer.gamer_version = str(newver)

            # No changes since 2.0.7... yet!
            elif compare_version(fileVer, (2, 0, 7)) >= 0:
                print("Migrating from v%s to v%s" % (str(fileVer), str(currVer)))
                scene.gamer.gamer_version = str(currVer)

        fileVer = literal_eval(scene.gamer.gamer_version)
        scene.gamer.versionerror = compare_version(fileVer, currVer)

    if scene.gamer.versionerror == 1:
        bpy.ops.gamer.prompt_update()


# VERSION UTILITY FUNCTIONS
def compare_version(v1, v2):
    """Compare version tuples

    Args:
        v1 (tuple): The version
        v2 (tuple): The other version

    Returns:
        int:  Return 1: v1 > v2, Return 0:  v1 == v2, Return -1: v1 <  v2
    """

    def cmp(a, b):
        """
        Compare a and b. Returns -1 if b > a, 1 if a > b, or 0 if a == b
        """
        return (a > b) - (a < b)

    return cmp(*zip(*map(lambda x, y: (x or 0, y or 0), map(int, v1), map(int, v2))))


classes = [
    GAMER_OT_prompt_update,
    GAMER_OT_prompt_old_version,
    GAMER_OT_update_to_2_0_1_from_v_0_1,
]


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(make_annotations(cls))


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(make_annotations(cls))



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# ***************************************************************************

# Report errors with the mesh.
# This file was copied from Blender 3D Print Addon

_data = []


def update(*args):
    _data[:] = args


def info():
    return tuple(_data)



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# ***************************************************************************

import bpy
from bpy.props import (
    BoolProperty,
    CollectionProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    IntProperty,
    IntVectorProperty,
    PointerProperty,
    StringProperty,
    BoolVectorProperty,
)
from bpy.app.handlers import persistent

from blendgamer.surfacemesh_ops import SurfaceMeshImprovementProperties
from blendgamer.versions import checkVersion, getGamerVersion
from blendgamer.meshstats import MeshQualityReportProperties
from blendgamer.tetrahedralization import GAMerTetrahedralizationPropertyGroup
from blendgamer.markers import GAMerBoundaryMarkersList
from blendgamer.curvatures import GAMerCurvaturesList
from blendgamer.util import UNSETID, make_annotations, get_material_by_bnd_id, getBndUnsetMat

import blendgamer.pygamer as pygamer

# python imports
import sys


@persistent
def gamer_load_post(dummy):
    """
    Initialize the GAMer addon.

    Checks that the addon properties are correctly initialized. Also that the
    plugin version matches the metadata layout stored in a blendfile. This
    function also sets a flag noting whether matplotlib is installed or not.
    """
    print(
        "Loading BlendGAMer v%s with PyGAMer %s"
        % (getGamerVersion(), pygamer.__version__())
    )
    scene = bpy.context.scene
    scene.gamer.check_for_matplotlib()
    if not scene.gamer.initialized:
        # print('Initializing GAMer Properties')
        scene.gamer.init_properties()
    else:
        # GAMer was previously initialized check version
        checkVersion()
        return

    bnd_unset_mat = getBndUnsetMat()


class GAMerAddonProperties(bpy.types.PropertyGroup):
    """
    Property group to store GAMer addon metadata
    """

    initialized = BoolProperty(name="GAMer Initialized", default=False)
    matplotlib_found = BoolProperty(name="Is matplotlib available", default=False)
    gamer_version = StringProperty(name="GAMer Version", default="0")
    boundary_id_counter = IntProperty(name="GAMer Boundary id Counter")
    versionerror = IntProperty(name="Version mismatch", default=0)

    surfmesh_improvement_properties = PointerProperty(
        type=SurfaceMeshImprovementProperties,
        name="GAMer Surface Mesh Improvement Properties",
    )

    mesh_quality_properties = PointerProperty(
        type=MeshQualityReportProperties, name="GAMer Mesh Quality Reporting"
    )

    tet_group = PointerProperty(
        type=GAMerTetrahedralizationPropertyGroup, name="GAMer Tetrahedralization"
    )

    def allocate_boundary_id(self):
        """Allocate the next available boundary ID.

        Returns
        -------
        int
            The boundary ID.
        """
        self.boundary_id_counter += 1
        return self.boundary_id_counter

    def init_properties(self):
        """Initialize BlendGAMer addon properties"""
        self.gamer_version = str(getGamerVersion())
        self.boundary_id_counter = 0  # Start counting at 0

        if "bnd_unset_mat" not in bpy.data.materials:
            bnd_unset_mat = bpy.data.materials.new("bnd_unset_mat")
            bnd_unset_mat.use_fake_user = True
            bnd_unset_mat.gamer.boundary_id = UNSETID
            self.initialized = True

    def check_for_matplotlib(self):
        """Check if matplotlib is available and set an internal flag."""
        import importlib.util

        mpl_spec = importlib.util.find_spec("matplotlib")
        self.matplotlib_found = mpl_spec is not None


class GAMerObjectProperties(bpy.types.PropertyGroup):
    """
    PropertyGroup of properties to link into Blender Objects
    """

    markers = PointerProperty(type=GAMerBoundaryMarkersList, name="Boundary Markers")
    curvatures = PointerProperty(type=GAMerCurvaturesList, name="Curvature Lists")


classes = [GAMerAddonProperties, GAMerObjectProperties]


def register():
    for cls in classes:
        bpy.utils.register_class(make_annotations(cls))


def unregister():
    for cls in reversed(classes):
        bpy.utils.unregister_class(make_annotations(cls))



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# ***************************************************************************

import bpy
import bmesh
from bpy.props import (
    BoolProperty,
    CollectionProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    IntProperty,
    IntVectorProperty,
    PointerProperty,
    StringProperty,
    BoolVectorProperty,
)

from blendgamer.util import *


# OPERATORS
# Object Boundary Marker Operators:
class GAMER_OT_add_boundary(bpy.types.Operator):
    bl_idname = "gamer.add_boundary"
    bl_label = "Add New Boundary"
    bl_description = "Add a new boundary to an object"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.object.gamer.markers.add_boundary(context)
        return {"FINISHED"}


class GAMER_OT_remove_boundary(bpy.types.Operator):
    bl_idname = "gamer.remove_boundary"
    bl_label = "Remove Boundary"
    bl_description = "Remove selected boundary from object"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.object.gamer.markers.remove_boundary(context)
        return {"FINISHED"}


class GAMER_OT_remove_all_boundaries(bpy.types.Operator):
    bl_idname = "gamer.remove_all_boundaries"
    bl_label = "Remove All Boundaries"
    bl_description = "Remove all boundaries from object"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.object.gamer.markers.remove_all_boundaries(context)
        return {"FINISHED"}


class GAMER_OT_assign_boundary_faces(bpy.types.Operator):
    bl_idname = "gamer.assign_boundary_faces"
    bl_label = "Assign Selected Faces To Boundary"
    bl_description = "Assign selected faces to boundary"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        bnd = context.object.gamer.markers.get_active_boundary()
        if bnd:
            bnd.assign_boundary_faces(context)
            return {"FINISHED"}
        self.report({"WARNING"}, "Cannot assign faces, no active boundary selected")
        return {"CANCELLED"}


class GAMER_OT_remove_boundary_faces(bpy.types.Operator):
    bl_idname = "gamer.remove_boundary_faces"
    bl_label = "Remove Selected Faces From Boundary"
    bl_description = "Remove selected faces from boundary"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        bnd = context.object.gamer.markers.get_active_boundary()
        if bnd:
            bnd.remove_boundary_faces(context)
            return {"FINISHED"}
        self.report({"WARNING"}, "Cannot remove faces, no active boundary selected")
        return {"CANCELLED"}


class GAMER_OT_select_boundary_faces(bpy.types.Operator):
    bl_idname = "gamer.select_boundary_faces"
    bl_label = "Select Faces of Selected Boundary"
    bl_description = "Select faces of selected boundary"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        bnd = context.object.gamer.markers.get_active_boundary()
        if bnd:
            bnd.select_boundary_faces(context)
            return {"FINISHED"}
        return {"CANCELLED"}


class GAMER_OT_deselect_boundary_faces(bpy.types.Operator):
    bl_idname = "gamer.deselect_boundary_faces"
    bl_label = "Deselect Faces of Selected Boundary"
    bl_description = "Deselect faces of selected boundary"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        bnd = context.object.gamer.markers.get_active_boundary()
        if bnd:
            bnd.deselect_boundary_faces(context)
            return {"FINISHED"}
        return {"CANCELLED"}


class GAMER_OT_select_all_boundary_faces(bpy.types.Operator):
    bl_idname = "gamer.select_all_boundary_faces"
    bl_label = "Select All Faces of Selected Boundary"
    bl_description = "Select all faces of selected boundary"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        for bnd in context.object.gamer.markers.boundary_list:
            bnd.select_boundary_faces(context)
        return {"FINISHED"}


class GAMER_OT_deselect_all_boundary_faces(bpy.types.Operator):
    bl_idname = "gamer.deselect_all_boundary_faces"
    bl_label = "Deselect all marked faces"
    bl_description = "Deselect all faces of selected boundary"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        for bnd in context.object.gamer.markers.boundary_list:
            bnd.deselect_boundary_faces(context)
        return {"FINISHED"}


class GAMerBoundaryMaterial(bpy.types.PropertyGroup):
    """
    Class for GAMer boundary material property group.
    """

    boundary_id = IntProperty(name="Boundary ID associated with material", default=-1)


class GAMerBoundaryMarker(bpy.types.PropertyGroup):
    """
    Class for GAMer boundary markers property group.
    """

    boundary_id = IntProperty(
        name="Boundary ID",
        description="Unique identifier of this boundary",
        min=-1,
        default=-1,
    )
    boundary_name = StringProperty(
        name="Boundary Name", description="Name of the boundary", default="Boundary"
    )
    marker = IntProperty(
        name="Marker Value",
        description="Marker value to associate with this boundary",
        default=1,
    )
    status = BoolProperty(name="Status", default=False)

    def init_boundary(self, context):
        """
        Initialize new boundary object
        """
        bnd_id = context.scene.gamer.allocate_boundary_id()
        bnd_name = "Boundary_%d" % (bnd_id)

        # Set ID prop key to bnd_id.
        # This should be considered immutable...
        self.name = str(bnd_id)

        obj = context.active_object

        with BMeshContext(obj) as bm:
            ml = getBMeshMarkerLayer

        # Get list of materials
        mats = bpy.data.materials
        bnd_unset_mat = getBndUnsetMat()
        # Ensure bnd_unset_mat is in object.material_slots
        if bnd_unset_mat.name not in obj.material_slots:
            # Add bnd_unset to material_slots...
            obj.data.materials.append(bnd_unset_mat)

        # Create a new boundary material. It shouldn't exist already
        # unless GAMerAddonProperties has been tampered with...
        bnd_mat_name = materialNamer(bnd_id)
        bnd_mat = bpy.data.materials.new(bnd_mat_name)
        # Ensure material is saved even if nothing is allocated to it
        bnd_mat.use_fake_user = True
        bnd_mat.gamer.boundary_id = bnd_id

        # Add new material to object material slots
        obj.data.materials.append(bnd_mat)

        self.boundary_id = bnd_id
        self.marker = bnd_id
        self.boundary_name = bnd_name

    def delete_boundary(self, context):
        """
        Remove boundary data from obj
        """
        obj = context.active_object
        mesh = obj.data

        # Clean up the materials
        mats = bpy.data.materials
        bnd_mat = get_material_by_bnd_id(self.boundary_id)

        with ObjectMode():
            # Material slots can only be removed in object mode!
            objmats = obj.data.materials
            # First remove all instances from slots
            while bnd_mat.name in objmats:
                idx = objmats.find(bnd_mat.name)
                if bpy.app.version < (2, 81, 0):
                    objmats.pop(index=idx, update_data=True)
                else:
                    objmats.pop(index=idx)

        # Remove the global material
        mats.remove(bnd_mat)

        # Assign unset material to members
        bnd_unset_mat = getBndUnsetMat()
        bnd_unset_mat_idx = obj.material_slots.find(bnd_unset_mat.name)
        # Link the material to the object if it's somehow missing
        if bnd_unset_mat_idx == -1:
            obj.data.materials.append(bnd_unset_mat)
            bnd_unset_mat_idx = obj.material_slots.find(bnd_unset_mat.name)

        with BMeshContext(obj) as bm:
            ml = getBMeshMarkerLayer(bm)

            for face in bm.faces:
                if face[ml] == self.boundary_id:
                    face[ml] = UNSETID
                    face.material_index = bnd_unset_mat_idx

    def assign_boundary_faces(self, context):
        """Assign boundary marker to selected faces

        Args:
            context (TYPE): Blender context
        """
        obj = context.active_object
        mesh = obj.data

        # Material to associate
        bnd_mat = get_material_by_bnd_id(self.boundary_id)
        matID = obj.material_slots.find(bnd_mat.name)
        # Link the material to the object if it's somehow missing
        if matID == -1:
            obj.data.materials.append(bnd_mat)
            matID = obj.material_slots.find(bnd_mat.name)

        if mesh.total_face_sel > 0:
            with BMeshContext(obj) as bm:
                ml = getBMeshMarkerLayer(bm)
                for face in bm.faces:
                    # Apply boundary marker label if selected
                    if face.select == True:
                        face[ml] = self.boundary_id
                        face.material_index = matID

    def repaint_boundary_faces(self, context):
        obj = context.active_object
        mats = bpy.data.materials

        # Add unset boundary to materials
        bnd_mat = getBndUnsetMat()
        matID = obj.material_slots.find(bnd_mat.name)
        # Link the material to the object if it's somehow missing
        if matID == -1:
            obj.data.materials.append(bnd_mat)
            matID = obj.material_slots.find(bnd_mat.name)

        # Material to associate
        bnd_mat = get_material_by_bnd_id(self.boundary_id)
        matID = obj.material_slots.find(bnd_mat.name)
        # Link the material to the object if it's somehow missing
        if matID == -1:
            obj.data.materials.append(bnd_mat)
            matID = obj.material_slots.find(bnd_mat.name)

        with BMeshContext(obj) as bm:
            ml = getBMeshMarkerLayer(bm)
            for face in bm.faces:
                # Apply boundary marker material
                if face[ml] == self.boundary_id:
                    face.material_index = matID

    def remove_boundary_faces(self, context):
        obj = context.active_object
        mesh = obj.data

        # Material to associate
        bnd_mat = getBndUnsetMat()
        matID = obj.material_slots.find(bnd_mat.name)
        # Link the material to the object if it's somehow missing
        if matID == -1:
            obj.data.materials.append(bnd_mat)
            matID = obj.material_slots.find(bnd_mat.name)

        if mesh.total_face_sel > 0:
            with BMeshContext(obj) as bm:
                ml = getBMeshMarkerLayer(bm)
                for face in bm.faces:
                    if face.select and face[ml] == self.boundary_id:
                        face[ml] = UNSETID
                        face.material_index = matID

    def select_boundary_faces(self, context):
        obj = context.active_object

        with BMeshContext(obj) as bm:
            ml = getBMeshMarkerLayer(bm)

            for face in bm.faces:
                if face[ml] == self.boundary_id:
                    face.select_set(True)

    def deselect_boundary_faces(self, context):
        obj = context.active_object

        with BMeshContext(obj) as bm:
            ml = getBMeshMarkerLayer(bm)

            for face in bm.faces:
                if face[ml] == self.boundary_id:
                    face.select_set(False)
            # bm.select_flush_mode()

    # TODO: (10) Enforce MCell boundary naming conventions
    # def check_boundary_name(self, bnd_name_list):
    #     status = ""
    #     # Check for duplicate boundary name
    #     if bnd_name_list.count(self.name) > 1:
    #         status = "Duplicate boundary: %s" % (self.name)

    #     # Check for illegal names (Starts with a letter. No special characters)
    #     bnd_filter = r"(^[A-Za-z]+[0-9A-Za-z_.]*$)"
    #     m = re.match(bnd_filter, self.name)
    #     if m is None:
    #         status = "Boundary name error: %s" % (self.name)

    #     self.status = status
    #     return


class GAMerBoundaryMarkersList(bpy.types.PropertyGroup):
    boundary_list = CollectionProperty(type=GAMerBoundaryMarker, name="Boundary List")
    active_bnd_index = IntProperty(name="Active Boundary Index", default=0)

    def get_active_boundary(self):
        """
        @brief      Gets the active boundary object
        @param      self
        @return     GAMerBoundaryMarkerPropertyGroup object of active boundary
        """
        bnd = None
        if len(self.boundary_list) > 0:
            bnd = self.boundary_list[self.active_bnd_index]
        return bnd

    def add_boundary(self, context):
        """
        Add a new boundary to the list of boundaries and set as the active
        boundary
        """
        # Create new GamerBoundaryMarker and initialize it
        new_bnd = self.boundary_list.add()
        new_bnd.init_boundary(context)

        # Set active index to new boundary
        idx = self.boundary_list.find(new_bnd.name)
        self.active_bnd_index = idx

    def repaint_boundaries(self, context):
        for bnd in self.boundary_list:
            bnd.repaint_boundary_faces(context)

    def remove_all_boundaries(self, context):
        for i in range(len(self.boundary_list)):
            # First remove boundary data from mesh:
            bnd = self.boundary_list[0]
            bnd.delete_boundary(context)

            # Now remove the boundary from the object
            self.boundary_list.remove(0)

        self.active_bnd_index = 0  # Restore original mode

    def remove_boundary(self, context):
        # First remove ID prop boundary data from object:
        bnd = self.get_active_boundary()
        if bnd:
            bnd.delete_boundary(context)

            # Now remove the RNA boundary from the object
            self.boundary_list.remove(self.active_bnd_index)
            self.active_bnd_index -= 1
            if self.active_bnd_index < 0:
                self.active_bnd_index = 0


classes = [
    GAMerBoundaryMarker,
    GAMerBoundaryMaterial,
    GAMerBoundaryMarkersList,
    GAMER_OT_add_boundary,
    GAMER_OT_remove_boundary,
    GAMER_OT_remove_all_boundaries,
    GAMER_OT_assign_boundary_faces,
    GAMER_OT_remove_boundary_faces,
    GAMER_OT_select_boundary_faces,
    GAMER_OT_deselect_boundary_faces,
    GAMER_OT_select_all_boundary_faces,
    GAMER_OT_deselect_all_boundary_faces,
]


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(make_annotations(cls))


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(make_annotations(cls))



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# **************************************************************************

import bpy
from bpy.props import (
    BoolProperty,
    CollectionProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    IntProperty,
    IntVectorProperty,
    PointerProperty,
    StringProperty,
    BoolVectorProperty,
)

import blendgamer.pygamer as g
from blendgamer.util import *

# python imports
import os
import numpy as np


class GAMER_OT_tet_domain_add(bpy.types.Operator):
    bl_idname = "gamer.tet_domain_add"
    bl_label = "Add a Tet Domain"
    bl_description = "Add a new tetrahedralization domain"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.scene.gamer.tet_group.add_tet_domain(self.report, context)
        return {"FINISHED"}


class GAMER_OT_tet_domain_remove(bpy.types.Operator):
    bl_idname = "gamer.tet_domain_remove"
    bl_label = "Remove a Tet Domain"
    bl_description = "Remove a tetrahedralization domain"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.scene.gamer.tet_group.remove_active_tet_domain(context)
        self.report({"INFO"}, "Deleted Active Tet Group")
        return {"FINISHED"}


class GAMER_OT_tet_domain_remove_all(bpy.types.Operator):
    bl_idname = "gamer.tet_domain_remove_all"
    bl_label = "Remove a Tet Domain"
    bl_description = "Remove a tetrahedralization domain"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        context.scene.gamer.tet_group.remove_all_tet_domains(context)
        self.report({"INFO"}, "Deleted Active Tet Groups")
        return {"FINISHED"}


class GAMER_OT_tetrahedralize(bpy.types.Operator):
    bl_idname = "gamer.tetrahedralize"
    bl_label = "Tetrahedralize"
    bl_description = "Tetrahedralize"
    bl_options = {"REGISTER"}

    def execute(self, context):
        context.scene.gamer.tet_group.tetrahedralize(context, self.report)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)


class GAMER_OT_surfaces_to_comsol(bpy.types.Operator):
    bl_idname = "gamer.surfaces_to_comsol"
    bl_label = "Export to Comsol"
    bl_description = "Export surface meshes to Comsol"
    bl_options = {"REGISTER"}

    def execute(self, context):
        context.scene.gamer.tet_group.surfaces_to_comsol(context, self.report)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)


class GAMER_OT_cleanup_domains(bpy.types.Operator):
    bl_idname = "gamer.cleanup_domains"
    bl_label = "Clean up domains"
    bl_description = "Clean up missing objects from domains"
    bl_options = {"REGISTER"}

    def execute(self, context):
        context.scene.gamer.tet_group.validate_domain_objects(context, self.report)
        return {"FINISHED"}

    def invoke(self, context, event):
        return self.execute(context)


class GAMerTetDomainPropertyGroup(bpy.types.PropertyGroup):
    # name = StringProperty()  # This is a reminder that "name" is already defined for all subclasses of PropertyGroup
    domain_id = IntProperty(name="ID", default=-1, description="Domain ID")
    # Deprecated in 2.0.7 in favor of pointer properties
    # object_name = StringProperty(name="ObjName", default="", description="Object Name")
    object_pointer = PointerProperty(
        type=bpy.types.Object, name="Test", description="Object"
    )
    marker = IntProperty(name="Marker", default=-1, description="Domain Marker Integer")
    is_hole = BoolProperty(
        name="Hole", default=False, description="Use this domain as a hole"
    )
    constrain_vol = BoolProperty(
        name="Constrain Volume", default=False, description="Constrain Volume"
    )
    vol_constraint = FloatProperty(
        name="Vol Constraint", default=10.0, description="Volume Constraint"
    )

    def get_name(self):
        if self.object_pointer is not None:
            return self.object_pointer.name
        return None

    def draw_layout(self, layout):
        row = layout.row()
        col = row.column()
        col.prop(self, "is_hole", text="Use Domain as a Hole")
        if not self.is_hole:
            col = row.column()
            col.prop(self, "marker")
            row = layout.row()
            col = row.column()
            col.prop(self, "constrain_vol")
            if self.constrain_vol:
                col = row.column()
                col.prop(self, "vol_constraint")

    def draw_item_in_row(self, row):
        name = self.get_name()
        if name is None:
            row.alert = True
            row.label(text="Object is missing")
            return
        elif bpy.context.scene.objects.get(self.object_pointer.name) == None:
            row.alert = True
            row.label(text="Object(%s) is not linked to scene" % (name))
            return

        col = row.column()
        col.label(text=name)
        col = row.column()
        col.label(text="Domain ID: " + str(self.domain_id))
        col = row.column()
        if self.is_hole:
            col.label(text="Hole")
        else:
            col.label(text="Domain Marker: " + str(self.marker))


class GAMerTetrahedralizationPropertyGroup(bpy.types.PropertyGroup):
    export_path = StringProperty(
        name="Export Directory",
        description="Path to directory where files will be created",
        default="//",
        maxlen=1024,
        subtype="DIR_PATH",
    )
    export_filebase = StringProperty(
        name="Filename",
        description="Base name of the files to export",
        default="gamertetmesh",
        maxlen=1024,
        subtype="FILE_NAME",
    )
    domain_list = CollectionProperty(
        type=GAMerTetDomainPropertyGroup, name="Domain List"
    )
    active_domain_index = IntProperty(name="Active Domain Index", default=0)
    next_id = IntProperty(
        name="Counter for Unique Domain IDs", default=1
    )  # Start ID's at 1 to confirm initialization

    show_settings = BoolProperty(
        name="Tetrahedralization Settings",
        default=False,
        description="Show more detailed settings",
    )

    min_dihedral = FloatProperty(
        name="Min Dihedral", default=10.0, description="Minimum Dihedral in Degrees"
    )
    max_aspect_ratio = FloatProperty(
        name="Max Aspect Ratio", default=1.3, description="Maximum Aspect Ratio"
    )

    ho_mesh = BoolProperty(
        name="Higher order mesh generation",
        default=False,
        description="Higher order mesh generation",
    )

    dolfin = BoolProperty(
        name="DOLFIN", default=False, description="Generate DOLFIN output"
    )
    # diffpack = BoolProperty(
    #         name="Diffpack", default=False,
    #         description="Generate Diffpack output")
    paraview = BoolProperty(
        name="Paraview", default=False, description="Generate Paraview output"
    )
    comsol = BoolProperty(
        name="Comsol (β)",
        default=False,
        description="(Untested) Generate Comsol mphtxt output",
    )

    status = StringProperty(name="status", default="")

    def add_tet_domain(self, report, context):
        """Add a new tet domain to the list of tet domains for each selected object"""
        # From the list of selected objects, only add MESH objects.
        objs = [obj for obj in context.selected_objects if obj.type == "MESH"]
        if len(objs) > 0:
            for obj in objs:
                # Check by name to see if it's already listed
                current_domain_names = [d.get_name() for d in self.domain_list]
                # print("Current domains = " + str(current_domain_names))
                if not (obj.name in current_domain_names):
                    new_id = (
                        self.allocate_available_id()
                    )  # Do this first to check for empty list before adding
                    new_dom = self.domain_list.add()
                    new_dom.name = str(new_id)
                    new_dom.domain_id = new_id
                    new_dom.marker = new_id
                    new_dom.object_pointer = obj
                    self.active_domain_index = len(self.domain_list) - 1
                    report({"INFO"}, "Added a new Tet domain")

    def remove_active_tet_domain(self, context):
        # print("Removing active Tet Domain")
        """Remove the active tet domain from the list of domains"""
        self.domain_list.remove(self.active_domain_index)
        self.active_domain_index -= 1
        if self.active_domain_index < 0:
            self.active_domain_index = 0

    def remove_domain_by_index(self, idx):
        """Remove the tet domain by index from the list of domains"""
        self.domain_list.remove(idx)
        if not (
            self.active_domain_index >= 0
            and self.active_domain_index < len(self.domain_list)
        ):
            self.active_domain_index = 0

    def remove_all_tet_domains(self, context):
        # print("Removing All Tet Domains")
        """Remove all tet domains from the list of domains"""
        while len(self.domain_list) > 0:
            self.domain_list.remove(0)
        self.active_domain_index = 0

    def allocate_available_id(self):
        """Return a unique domain ID for a new domain"""
        # print ("Next ID is " + str(self.next_id))
        if len(self.domain_list) <= 0:
            # Reset the ID to 1 when there are no more molecules
            self.next_id = 1
        self.next_id += 1
        return self.next_id - 1

    def validate_domain_objects(self, context, report):
        for i in range(len(self.domain_list) - 1, -1, -1):
            d = self.domain_list[i]
            if d.object_pointer is None:
                self.remove_domain_by_index(i)
            elif context.scene.objects.get(d.object_pointer.name) is None:
                bpy.data.objects.remove(d.object_pointer)
                self.remove_domain_by_index(i)

    def surfaces_to_comsol(self, context, report):
        self.validate_domain_objects(context, report)

        filename = self.export_path + self.export_filebase
        filename = bpy.path.abspath(filename)
        gmeshes = list()

        current_domain_names = [d.get_name() for d in self.domain_list]
        print("Current domains = " + str(current_domain_names))

        for d in self.domain_list:
            obj = d.object_pointer
            gmesh = blender_to_gamer(obj=obj, map_boundaries=True)
            if not gmesh:
                print("blender_to_gamer returned a gmesh of None")
            else:
                print(
                    "Mesh %s: num verts: %d numfaces: %d"
                    % (obj.name, gmesh.nVertices, gmesh.nFaces)
                )
                # Set the domain data on the SurfaceMesh these are the per/domain items as_hole, marker, and volume constraints
                print("Closed: %d; Marker: %d" % (d.is_hole, d.marker))

                globalInfo = gmesh.getRoot()
                globalInfo.ishole = d.is_hole
                globalInfo.marker = d.marker
                globalInfo.useVolumeConstraint = d.constrain_vol
                globalInfo.volumeConstraint = d.vol_constraint

                # Write surface mesh to file for debug
                # g.writeOFF("surfmesh_%s.off"%(obj_name), gmesh)

                # Add the mesh
                gmeshes.append(gmesh)
        if len(gmeshes) > 0:
            report(
                {"INFO"},
                "Writing surface meshes to Comsol file: " + filename + ".mphtxt",
            )
            g.writeComsol(filename + ".mphtxt", gmeshes)

    def tetrahedralize(self, context, report):
        self.validate_domain_objects(context, report)
        print("######################## Begin Tetrahedralize ########################")

        # filename = self.tet_path
        filename = self.export_path + self.export_filebase
        filename = bpy.path.abspath(filename)
        if not (self.dolfin or self.paraview or self.comsol):
            self.status = (
                "Please select an output format in Tetrahedralization Settings"
            )
            print(self.status)
        else:
            self.status = ""
            mesh_formats = []

            if self.dolfin:
                mesh_formats.append("dolfin")
            if self.paraview:
                mesh_formats.append("paraview")
            if self.comsol:
                mesh_formats.append("comsol")

            # Vector of SurfaceMeshes
            # gmeshes = g.VectorSM()
            gmeshes = list()
            # for (obj_name, tet_domain) in [
            #     (d.object_name, d) for d in self.domain_list
            # ]:
            #     print("obj_name = " + obj_name + ", tet_domain = " + str(tet_domain))

            current_domain_names = [d.get_name() for d in self.domain_list]
            print("Current domains = " + str(current_domain_names))

            for d in self.domain_list:
                obj = d.object_pointer
                gmesh = blender_to_gamer(obj=obj, map_boundaries=True)

                if not gmesh:
                    print("blender_to_gamer returned a gmesh of None")
                else:
                    print(
                        "Mesh %s: num verts: %d numfaces: %d"
                        % (obj.name, gmesh.nVertices, gmesh.nFaces)
                    )
                    # Set the domain data on the SurfaceMesh these are the per/domain items as_hole, marker, and volume constraints
                    print("Closed: %d; Marker: %d" % (d.is_hole, d.marker))

                    globalInfo = gmesh.getRoot()
                    globalInfo.ishole = d.is_hole
                    globalInfo.marker = d.marker
                    globalInfo.useVolumeConstraint = d.constrain_vol
                    globalInfo.volumeConstraint = d.vol_constraint

                    # Write surface mesh to file for debug
                    # g.writeOFF("surfmesh_%s.off"%(obj_name), gmesh)

                    # Add the mesh
                    gmeshes.append(gmesh)

            # Tetrahedralize mesh
            if len(gmeshes) > 0:

                quality_str = "q%.4f/%.4fO8/7AYVCa" % (
                    self.max_aspect_ratio,
                    self.min_dihedral,
                )

                # if self.constrain_vol:
                #     # a%.8f volume constraint..
                #     quality_str.append('a%.8f', self.vol_constraint)
                quality_str += "o2" if self.ho_mesh else ""

                print("========================================")
                print("TetGen quality string: " + quality_str)
                print("========================================")
                # Do the tetrahedralization
                tetmesh = g.makeTetMesh(gmeshes, quality_str)

                # for i in range(0,5):
                #     print("Laplacian smooth iteration.")
                #     g.smoothMesh(tetmesh)

                # Store mesh to files
                tetmesh_formats = ["dolfin", "paraview", "comsol"]
                tetmesh_suffices = [".xml", ".vtk", ".mphtxt"]

                for fmt in mesh_formats:
                    try:
                        suffix = tetmesh_suffices[tetmesh_formats.index(fmt)]
                        report(
                            {"INFO"},
                            "Writing to " + fmt + " file: " + filename + suffix,
                        )
                        if fmt == "dolfin":
                            g.writeDolfin(filename + suffix, tetmesh)
                        if fmt == "paraview":
                            g.writeVTK(filename + suffix, tetmesh)
                        if fmt == "comsol":
                            g.writeComsol(filename + suffix, tetmesh)
                    except Exception as ex:
                        report({"ERROR"}, str(ex))

        print("######################## End Tetrahedralize ########################")


classes = [
    GAMER_OT_tet_domain_add,
    GAMER_OT_tet_domain_remove,
    GAMER_OT_tet_domain_remove_all,
    GAMER_OT_tetrahedralize,
    GAMER_OT_surfaces_to_comsol,
    GAMER_OT_cleanup_domains,
    GAMerTetDomainPropertyGroup,
    GAMerTetrahedralizationPropertyGroup,
]


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(make_annotations(cls))


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(make_annotations(cls))



import pytest
import pygamer
import pygamer.surfacemesh as sm
import pygamer.tetmesh as tm
import math

class TestVertex(object):
    def test_constructors(self):
        vertex = sm.Vertex()
        assert vertex[0] == 0
        assert vertex[1] == 0
        assert vertex[2] == 0
        assert vertex.marker == -1
        assert vertex.selected == False

        vertex = sm.Vertex(1,2,3,25,True)
        assert vertex[0] == 1
        assert vertex[1] == 2
        assert vertex[2] == 3
        assert vertex.marker == 25
        assert vertex.selected == True

    def test_assignment(self):
        vertex = sm.Vertex()

        vertex[0] = 5
        vertex[1] = 8
        vertex[2] = 10
        vertex.marker = 25
        vertex.selected = True

        assert vertex[0] == 5
        assert vertex[1] == 8
        assert vertex[2] == 10
        assert vertex.marker == 25
        assert vertex.selected == True

        with pytest.raises(IndexError):
            print(vertex[3])

class TestEdge(object):
    def test_constructors(self):
        edge = sm.Edge()
        assert edge.selected == False

        edge.selected = True
        assert edge.selected == True

        edge = sm.Edge(True)
        assert edge.selected == True


class TestFace(object):
    def test_constructors(self):
        face = sm.Face()

        assert face.orientation == 0
        assert face.marker == -1
        assert face.selected == False


        face = sm.Face(3, True)

        assert face.orientation == 0
        assert face.marker == 3
        assert face.selected == True

        face = sm.Face(1,2,True)

        assert face.orientation == 1
        assert face.marker == 2
        assert face.selected == True

    def test_assignment(self):
        face = sm.Face()
        face.orientation = 1
        face.marker = 10
        face.selected = True

        assert face.orientation == 1
        assert face.marker == 10
        assert face.selected == True

class TestSurfaceMesh(object):
    def test_insertion(self):
        mesh = sm.SurfaceMesh()
        mesh.insertFace([1,2,3])
        mesh.insertFace([2,3,4])
        assert mesh.nVertices == 4
        assert mesh.nEdges == 5
        assert mesh.nFaces == 2

        mesh.insertEdge([5,6])
        assert mesh.nVertices == 6
        assert mesh.nEdges == 6
        assert mesh.nFaces == 2

        mesh.insertVertex([7])
        assert mesh.nVertices == 7
        assert mesh.nEdges == 6
        assert mesh.nFaces == 2

        mesh.removeVertex([1])
        assert mesh.nVertices == 6
        assert mesh.nEdges == 4
        assert mesh.nFaces == 1


    def test_assignment(self):
        mesh = sm.SurfaceMesh()
        mesh.insertFace([1,2,3])

        fid = mesh.getFace([1,2,3])
        data = fid.data()

        assert data.orientation == 0
        assert data.marker == 0
        assert data.selected == False

        data.orientation = 1
        data.marker = 2
        data.selected = True

        fid = mesh.getFace([1,2,3])
        data = fid.data()

        assert data.orientation == 1
        assert data.marker == 2
        assert data.selected == True

        metadata = mesh.getRoot()
        assert metadata.marker == -1
        assert math.isclose(metadata.volumeConstraint, -1, rel_tol=1e-5) == True
        assert metadata.useVolumeConstraint == False
        assert metadata.ishole == False

        metadata.marker = 5
        metadata.volumeConstraint = 3.14159
        metadata.useVolumeConstraint = True
        metadata.ishole = True

        metadata = mesh.getRoot()
        assert metadata.marker == 5
        assert math.isclose(metadata.volumeConstraint, 3.14159, rel_tol=1e-5) == True
        assert metadata.useVolumeConstraint == True
        assert metadata.ishole == True

        vid = mesh.getVertex([1])
        data = vid.data()
        print(data.position)
        assert data.marker == -1
        assert data.selected == False




# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# ***************************************************************************

import bpy
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from blendgamer.util import *

# Recommended backend if a non-interactive backend is the default
# $ pip install pyqt5
# mpl.use('QT5Agg')

mpl.rcParams["pdf.fonttype"] = 42
mpl.rcParams["ps.fonttype"] = 42

# Mapping colormap_enums.colormap_enums to values
colormapDict = {
    "VIRIDIS": plt.cm.viridis,
    "PRGN": plt.cm.PRGn,
}


class DivergingNorm(mpl.colors.Normalize):
    def __init__(self, vcenter, vmin=None, vmax=None):
        """
        Normalize data with a set center.

        Useful when mapping data with an unequal rates of change around a
        conceptual center, e.g., data that range from -2 to 4, with 0 as
        the midpoint.

        Parameters
        ----------
        vcenter : float
            The data value that defines ``0.5`` in the normalization.
        vmin : float, optional
            The data value that defines ``0.0`` in the normalization.
            Defaults to the min value of the dataset.
        vmax : float, optional
            The data value that defines ``1.0`` in the normalization.
            Defaults to the the max value of the dataset.

        Examples
        --------
        This maps data value -4000 to 0., 0 to 0.5, and +10000 to 1.0; data
        between is linearly interpolated::

            >>> import matplotlib.colors as mcolors
            >>> offset = mcolors.DivergingNorm(vmin=-4000.,
                                               vcenter=0., vmax=10000)
            >>> data = [-4000., -2000., 0., 2500., 5000., 7500., 10000.]
            >>> offset(data)
            array([0., 0.25, 0.5, 0.625, 0.75, 0.875, 1.0])
        """

        self.vcenter = vcenter
        self.vmin = vmin
        self.vmax = vmax
        if not (vcenter and vmin and vmax) and (vcenter >= vmax or vcenter <= vmin):
            raise ValueError(
                "vmin(%f), vcenter(%f), and vmax(%f) must be in "
                "ascending order" % (vmin, vcenter, vmax)
            )

    def autoscale_None(self, A):
        """
        Get vmin and vmax, and then clip at vcenter
        """
        super().autoscale_None(A)
        if self.vmin > self.vcenter:
            self.vmin = self.vcenter
        if self.vmax < self.vcenter:
            self.vmax = self.vcenter

    def __call__(self, value, clip=False):
        """
        Map value to the interval [0, 1].
        """
        result, is_scalar = self.process_value(value)
        self.autoscale_None(result)  # sets self.vmin, self.vmax if None

        if not self.vmin <= self.vcenter <= self.vmax:
            raise ValueError("vmin, vcenter, vmax must increase monotonically")
        result = np.ma.masked_array(
            np.interp(result, [self.vmin, self.vcenter, self.vmax], [0, 0.5, 1.0]),
            mask=np.ma.getmask(result),
        )
        if is_scalar:
            result = np.atleast_1d(result)[0]
        return result


def curveToData(crv, context):
    """
    Helper function to take a curvature object and return smoothed data.

    Parameters
    ----------
    crv : Curvature object
        RNA object with curvature meta information
    context : Context of the scene
        Description

    Returns
    -------
    numpy.ndarray
        Array of data values
    """
    obj = getActiveMeshObject()
    bm = bmesh_from_object(obj)

    with ObjectMode():
        layer = getCurvatureLayer(obj, crv.algorithm, crv.curvatureType)

    data = np.zeros(len(obj.data.vertices), dtype=np.float)
    # Copy curvatures over
    layer.foreach_get("value", data)

    if crv.curveIter > 0:
        for i in range(0, crv.curveIter):
            tmp = np.zeros(len(obj.data.vertices), dtype=np.float)
            for v in bm.verts:
                count = 1
                tmp[v.index] = data[v.index]

                for e in v.link_edges:
                    v_other = e.other_vert(v)

                    tmp[v.index] += data[v_other.index]
                    count += 1
                tmp[v.index] /= count
            data = np.array(tmp, copy=True)
    return data


def dataToVertexColor(crv, context, showplot=False, saveplot=False):
    """
    Convert curvature object to colormap

    Parameters
    ----------
    crv : TYPE
        Description
    context : TYPE
        Description
    showplot : bool, optional
        Description
    saveplot : bool, optional
        Description
    """

    data = curveToData(crv, context)
    cmap = colormapDict[crv.colormap]
    file_prefix = "%s_%s_m%dM%dI%dmx%0.2f%s" % (
        bpy.path.basename(bpy.context.blend_data.filepath).split(".")[0],
        context.object.name,
        crv.minCurve,
        crv.maxCurve,
        crv.curveIter,
        crv.mixpoint,
        crv.curvatureType,
    )

    fig = plt.figure(figsize=(8, 5))
    ax = fig.add_axes([0.1, 0.05, 0.6, 0.9])

    ## This code helps make the plots easier to read...
    # tmin = np.percentile(data,1)
    # tmax = np.percentile(data,99)
    # data[data < tmin] =tmin
    # data[data > tmax] =tmax
    #
    tmin = crv.minCurve
    tmax = crv.maxCurve
    amin = np.amin(data)
    amax = np.amax(data)
    amean = np.mean(data)
    amedian = np.median(data)

    if showplot:
        ax.hist(data, bins="auto")
        ax.set_title("%s Distribution" % (file_prefix))
        ax.axvline(amin, color="r", linestyle="dashed", linewidth=1)
        ax.axvline(amax, color="r", linestyle="dashed", linewidth=1)

    extend = "neither"
    # the tmin/tmax values are percentiles instead
    if crv.limitsArePercentiles:
        # print("Using min/max values as percentiles!")
        if tmin < 0 or tmin >= 100:
            print("Minimum percentile must be 0<=x<100. Setting to 0")
            lowerPercentile = 0
        else:
            lowerPercentile = int(tmin)

        if tmax <= 0 or tmax > 100:
            print("Maximum percentile must be 0<x<=100. Setting to 100")
            upperPercentile = 100
        else:
            upperPercentile = int(tmax)

        tmin = np.percentile(data, lowerPercentile)
        tmax = np.percentile(data, upperPercentile)
        print(
            "Data truncated at %f and %f percentiles\n"
            % (lowerPercentile, upperPercentile)
        )

    if showplot:
        ax.axvline(tmin, color="g", linestyle="dashed", linewidth=2)
        ax.axvline(tmax, color="g", linestyle="dashed", linewidth=2)

        if tmin > amin and tmax < amax:
            extend = "both"
        elif tmin > amin:
            extend = "min"
        elif tmax < amax:
            extend = "max"

    data[data < tmin] = tmin
    data[data > tmax] = tmax

    amin = np.amin(data)
    amax = np.amax(data)

    # if amin > tmin:
    #     amin = tmin
    # if amax < tmax:
    #     amax = tmax

    # Construct the norm and colorbar
    if amin < 0 and amax > 0:
        norm = DivergingNorm(vmin=amin, vcenter=0, vmax=amax)
        # Python 3.5 matplotlib may not support?
        # norm = mpl.colors.DivergingNorm(vmin=amin, vcenter=0, vmax=amax)
        colors_neg = cmap(np.linspace(0, crv.mixpoint, 256))
        colors_pos = cmap(np.linspace(crv.mixpoint, 1, 256))

        all_colors = np.vstack((colors_neg, colors_pos))
        curvature_map = mpl.colors.LinearSegmentedColormap.from_list(
            "curvature_map", all_colors
        )
    else:
        norm = mpl.colors.Normalize(vmin=amin, vmax=amax)
        curvature_map = cmap

    # Map values to colors and add vertex color layer
    colors = curvature_map(norm(data))

    # Create view without alpha channel if Blender < 2.80
    if bpy.app.version < (2, 80, 0):
        colors = colors[:, :3]

    mesh = bpy.context.object.data
    vlayer = "%s%s" % (crv.algorithm, crv.curvatureType)

    if vlayer not in mesh.vertex_colors:
        if len(mesh.vertex_colors) == 8:
            raise RuntimeError(
                "Maximum of 8 vertex Layers reached cannot create a new layer. Please delete a layer to continue."
            )
        mesh.vertex_colors.new(name=vlayer)

    color_layer = mesh.vertex_colors[vlayer]
    mesh.vertex_colors[vlayer].active = True

    mloops = np.zeros((len(mesh.loops)), dtype=np.int)
    mesh.loops.foreach_get("vertex_index", mloops)
    color_layer.data.foreach_set("color", colors[mloops].flatten())

    # Add axis for colorbar and plot it
    ax = fig.add_axes([0.75, 0.05, 0.05, 0.9])
    cb = mpl.colorbar.ColorbarBase(
        ax, cmap=curvature_map, norm=norm, orientation="vertical"
    )

    ticks = cb.get_ticks()
    ticks.sort()

    if amin != ticks[0]:
        ticks = np.insert(ticks, 0, amin)
    if amax != ticks[-1]:
        ticks = np.append(ticks, amax)
    cb.set_ticks(ticks)

    ticklabels = [r"{:0.1f}".format(tick) for tick in ticks]

    if extend == "neither":
        pass
    elif extend == "both":
        ticklabels[0] = "< " + ticklabels[0]
        ticklabels[-1] = "> " + ticklabels[-1]
    elif extend == "max":
        ticklabels[-1] = "> " + ticklabels[-1]
    elif extend == "min":
        ticklabels[0] = "< " + ticklabels[0]
    cb.set_ticklabels(ticklabels)
    cb.ax.tick_params(labelsize=14)
    cb.set_label("%s [$\mu m^{-1}$]" % (vlayer), size=16)

    if saveplot:
        plt.savefig(file_prefix + ".pdf", format="pdf")
    if showplot:
        plt.show()
    plt.close()


def differencePlotter(context, difftype="K1"):
    obj = getActiveMeshObject()
    bm = bmesh_from_object(obj)

    with ObjectMode():
        mdsb = getCurvatureLayer(obj, "MDSB", difftype)
        jets = getCurvatureLayer(obj, "JETS", difftype)

        mdsb_data = np.zeros(len(obj.data.vertices), dtype=np.float)
        mdsb.foreach_get("value", mdsb_data)

        jets_data = np.zeros(len(obj.data.vertices), dtype=np.float)
        jets.foreach_get("value", jets_data)

        tmpmdsb = np.zeros(len(obj.data.vertices), dtype=np.float)
        tmpjets = np.zeros(len(obj.data.vertices), dtype=np.float)
        for v in bm.verts:
            count = 1
            tmpmdsb[v.index] = mdsb_data[v.index]
            tmpjets[v.index] = jets_data[v.index]

            for e in v.link_edges:
                v_other = e.other_vert(v)

                tmpmdsb[v.index] += mdsb_data[v_other.index]
                tmpjets[v.index] += jets_data[v_other.index]
                count += 1
            tmpmdsb[v.index] /= count
            tmpjets[v.index] /= count
        mdsb_data = np.array(tmpmdsb, copy=True)
        jets_data = np.array(tmpjets, copy=True)
    data = mdsb_data - jets_data

    cmap = colormapDict["PRGN"]
    file_prefix = "%s_difference" % (difftype)

    fig = plt.figure(figsize=(8, 5))
    ax = fig.add_axes([0.1, 0.05, 0.6, 0.9])

    amin = np.amin(data)
    amax = np.amax(data)
    amean = np.mean(data)
    amedian = np.median(data)

    plt.hist(data, bins="auto")
    plt.title("%s Distribution" % (file_prefix))
    plt.axvline(amin, color="r", linestyle="dashed", linewidth=1)
    plt.axvline(amax, color="r", linestyle="dashed", linewidth=1)

    # Save full data
    np.savez(context.object.name + "difference" + difftype + ".npz", data)
    extend = "neither"
    # tmin = amin
    # tmax = amax
    tmin = np.percentile(data, 1)
    tmax = np.percentile(data, 99)

    ax.axvline(tmin, color="g", linestyle="dashed", linewidth=2)
    ax.axvline(tmax, color="g", linestyle="dashed", linewidth=2)

    if tmin > amin and tmax < amax:
        extend = "both"
    elif tmin > amin:
        extend = "min"
    elif tmax < amax:
        extend = "max"

    data[data < tmin] = tmin
    data[data > tmax] = tmax

    amin = np.amin(data)
    amax = np.amax(data)

    # Construct the norm and colorbar
    if amin < 0 and amax > 0:
        norm = DivergingNorm(vmin=amin, vcenter=0, vmax=amax)
        # Python 3.5 matplotlib may not support?
        # norm = mpl.colors.DivergingNorm(vmin=amin, vcenter=0, vmax=amax)
        colors_neg = cmap(np.linspace(0, 0.5, 256))
        colors_pos = cmap(np.linspace(0.5, 1, 256))

        all_colors = np.vstack((colors_neg, colors_pos))
        curvature_map = mpl.colors.LinearSegmentedColormap.from_list(
            "curvature_map", all_colors
        )
    else:
        norm = mpl.colors.Normalize(vmin=amin, vmax=amax)
        curvature_map = cmap

    # Map values to colors and add vertex color layer
    colors = curvature_map(norm(data))

    # Create view without alpha channel if Blender < 2.80
    if bpy.app.version < (2, 80, 0):
        colors = colors[:, :3]

    mesh = bpy.context.object.data
    vlayer = "%s_diff" % (difftype)

    if vlayer not in mesh.vertex_colors:
        mesh.vertex_colors.new(name=vlayer)

    color_layer = mesh.vertex_colors[vlayer]
    mesh.vertex_colors[vlayer].active = True

    mloops = np.zeros((len(mesh.loops)), dtype=np.int)
    mesh.loops.foreach_get("vertex_index", mloops)
    color_layer.data.foreach_set("color", colors[mloops].flatten())

    # Add axis for colorbar and plot it
    ax = fig.add_axes([0.75, 0.05, 0.05, 0.9])
    cb = mpl.colorbar.ColorbarBase(
        ax, cmap=curvature_map, norm=norm, orientation="vertical"
    )

    ticks = cb.get_ticks()
    ticks.sort()

    if amin != ticks[0]:
        ticks = np.insert(ticks, 0, amin)
    if amax != ticks[-1]:
        ticks = np.append(ticks, amax)
    cb.set_ticks(ticks)

    ticklabels = [r"{:0.1f}".format(tick) for tick in ticks]

    # extend = 'neither'
    if extend == "neither":
        pass
    elif extend == "both":
        ticklabels[0] = "< " + ticklabels[0]
        ticklabels[-1] = "> " + ticklabels[-1]
    elif extend == "max":
        ticklabels[-1] = "> " + ticklabels[-1]
    elif extend == "min":
        ticklabels[0] = "< " + ticklabels[0]
    cb.set_ticklabels(ticklabels)
    cb.ax.tick_params(labelsize=14)
    cb.set_label("%s [$\mu m^{-1}$]" % (vlayer), size=16)

    plt.savefig(context.object.name + "difference" + difftype + ".pdf", format="pdf")
    # plt.show()
    plt.close()


def eng_notation(x, pos):
    num, power = "{:.1e}".format(x).split("e")
    power = int(power)
    return r"${} \times 10^{{{}}}$".format(num, power)



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# ***************************************************************************

import bpy
import bmesh
from bpy.props import (
    BoolProperty,
    CollectionProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    IntProperty,
    IntVectorProperty,
    PointerProperty,
    StringProperty,
    BoolVectorProperty,
)
from blendgamer.colormap_enums import colormap_enums

import importlib.util

mpl_spec = importlib.util.find_spec("matplotlib")
mpl_found = mpl_spec is not None

if mpl_found:
    from blendgamer.colormap import dataToVertexColor, differencePlotter

from blendgamer.util import *

curvatureTypeEnums = [
    ("K1", "k1", "First principle curvature"),
    ("K2", "k2", "Second principle curvature"),
    ("KG", "kg", "Gaussian curvature"),
    ("KH", "kh", "Mean curvature"),
]

curvatureCalcEnums = [
    ("MDSB", "MDSB", "Meyer, Desbrun, Schroder, Barr Algorithm"),
    ("JETS", "Jet Fitting", "Cazal and Pouget Jet Fitting"),
]
curvatureCalcDict = {
    "MDSB": "curvatureViaMDSB",
    "JETS": "curvatureViaJets",
}


class GAMER_OT_remove_curvature(bpy.types.Operator):
    bl_idname = "gamer.remove_curvature"
    bl_label = "Remove Curvature"
    bl_description = "Remove selected curvature from object"
    bl_options = {"REGISTER"}

    def execute(self, context):
        try:
            context.object.gamer.curvatures.remove_curvature(context)
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class GAMER_OT_remove_all_curvatures(bpy.types.Operator):
    bl_idname = "gamer.remove_all_curvatures"
    bl_label = "Remove All Curvatures"
    bl_description = "Remove all curvatures from object"
    bl_options = {"REGISTER"}

    def execute(self, context):
        try:
            context.object.gamer.curvatures.remove_all_curvatures(context)
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class GAMER_OT_plot_curvature(bpy.types.Operator):
    bl_idname = "gamer.plot_curvature"
    bl_label = "Plot the selected curvature"
    bl_description = "Plot the selected curvature as a heatmap and generate a corresponding vertex color layer"
    bl_options = {"REGISTER"}

    def execute(self, context):
        try:
            context.object.gamer.curvatures.plot_curvature(context)
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class GAMER_OT_plot_all_curvatures(bpy.types.Operator):
    bl_idname = "gamer.plot_all_curvatures"
    bl_label = "Plot all curvatures"
    bl_description = (
        "Plot all computed curvatures as a heatmap and generate vertex color layers"
    )
    bl_options = {"REGISTER"}

    def execute(self, context):
        try:
            context.object.gamer.curvatures.plot_all_curvatures(context)
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


# class GAMER_OT_plot_differences(bpy.types.Operator):
#     bl_idname       = "gamer.plot_differences"
#     bl_label        = "Plot Differences"
#     bl_description  = "Plot differences between MDSB and JETS"
#     bl_options      = {'REGISTER'}

#     def execute(self, context):
#         try:
#             context.object.gamer.curvatures.plot_differences(context)
#             return {'FINISHED'}
#         except Exception as e:
#             self.report({'ERROR'}, str(e))
#             return {'CANCELLED'}


class GAMER_OT_compute_curvatures(bpy.types.Operator):
    bl_idname = "gamer.compute_curvatures"
    bl_label = "Compute Curvatures"
    bl_description = "Compute curvatures"
    bl_options = {"REGISTER"}

    def execute(self, context):
        try:
            obj = getActiveMeshObject()
            curvatures = obj.gamer.curvatures
            algorithm = obj.gamer.curvatures.algorithm

            gmesh = blender_to_gamer()
            kh, kg, k1, k2 = getattr(gmesh, curvatureCalcDict[algorithm])(0)

            with ObjectMode():
                ml = getCurvatureLayer(obj, algorithm, "K1")
                for i in range(0, len(k1)):
                    ml[i].value = k1[i]
                curvatures.add_curvature(context, "K1")

                ml = getCurvatureLayer(obj, algorithm, "K2")
                for i in range(0, len(k1)):
                    ml[i].value = k2[i]
                curvatures.add_curvature(context, "K2")

                ml = getCurvatureLayer(obj, algorithm, "KG")
                for i in range(0, len(k1)):
                    ml[i].value = kg[i]
                curvatures.add_curvature(context, "KG")

                ml = getCurvatureLayer(obj, algorithm, "KH")
                for i in range(0, len(k1)):
                    ml[i].value = kh[i]
                curvatures.add_curvature(context, "KH")

            # Explicitly free curvature arrays
            del kh
            del kg
            del k1
            del k2

            self.report({"INFO"}, "GAMer: Compute Curvatures complete")
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class GAMerCurvatureItem(bpy.types.PropertyGroup):
    curvatureType = EnumProperty(
        name="Curvature Type", description="Which curvature?", items=curvatureTypeEnums
    )
    algorithm = EnumProperty(
        name="Curvature algorithm used",
        description="Which algorithm was used?",
        items=curvatureCalcEnums,
    )

    ## Addition metadata for saving plot states...
    minCurve = FloatProperty(
        name="Minimum curvature",
        default=0,
        description="Lower bound percentile truncation",
    )

    maxCurve = FloatProperty(
        name="Maximum curvature",
        default=100,
        description="Upper bound percentile truncation",
    )

    curveIter = IntProperty(
        name="Smooth Curvature Iterations",
        min=0,
        default=0,
        description="How many iterations of curvature smoothing?",
    )

    limitsArePercentiles = BoolProperty(
        name="Treat Min/Max as percentiles",
        default=True,
        description="Treat min and max as percentiles?",
    )

    mixpoint = FloatProperty(
        name="Color mixing point",
        default=0.5,
        min=0,
        max=1,
        description="Value for color mixing",
    )

    colormap = EnumProperty(
        name="Colormap colors", description="Colormap used", items=colormap_enums
    )


class GAMerCurvaturesList(bpy.types.PropertyGroup):
    algorithm = EnumProperty(
        name="Curvature Algorithm",
        description="Which algorithm was used?",
        items=curvatureCalcEnums,
    )

    curvature_list = CollectionProperty(
        type=GAMerCurvatureItem, name="List of computed curvatures"
    )

    active_index = IntProperty(name="Active Index", default=0)

    showplots = BoolProperty(
        name="Show plot", default=False, description="Display the plots"
    )

    saveplots = BoolProperty(
        name="Save plots", default=False, description="Save the generated plots"
    )

    def add_curvature(self, context, curvatureType):
        if len(self.curvature_list) > 0:
            for c in self.curvature_list:
                if c.algorithm == self.algorithm:
                    if c.curvatureType == curvatureType:
                        return False
        new_curve = self.curvature_list.add()
        new_curve.curvatureType = curvatureType
        new_curve.algorithm = self.algorithm

    def remove_curvature(self, context):
        crv = self.get_active_index()
        if crv:
            obj = getActiveMeshObject()

            with BMeshContext(obj) as bm:
                name = "%s%s" % (crv.algorithm, crv.curvatureType)
                layer = bm.verts.layers.float.get(name)
                bm.verts.layers.float.remove(layer)

            self.curvature_list.remove(self.active_index)
            self.active_index -= 1
            if self.active_index < 0:
                self.active_index = 0

    def remove_all_curvatures(self, context):
        obj = getActiveMeshObject()

        with BMeshContext(obj) as bm:
            for i in range(len(self.curvature_list)):
                crv = self.curvature_list[0]

                name = "%s%s" % (crv.algorithm, crv.curvatureType)
                layer = bm.verts.layers.float.get(name)
                bm.verts.layers.float.remove(layer)
                self.curvature_list.remove(0)

        self.active_index = 0

    def get_active_index(self):
        idx = None
        if len(self.curvature_list) > 0:
            idx = self.curvature_list[self.active_index]
        return idx

    if mpl_found:

        def plot_curvature(self, context):
            """
            Wrapper function to avoid importing lots of matplotlib files here.
            """
            crv = self.get_active_index()
            dataToVertexColor(
                crv, context, showplot=self.showplots, saveplot=self.saveplots
            )

        def plot_all_curvatures(self, context):
            """
            Wrapper function to plot all curvatures
            """
            for crv in self.curvature_list:
                dataToVertexColor(
                    crv, context, showplot=self.showplots, saveplot=self.saveplots
                )

        # def plot_differences(self, context):
        #     differencePlotter(context, 'K1')
        #     differencePlotter(context, 'K2')
        #     # differencePlotter(context, 'KG')
        #     # differencePlotter(context, 'KH')


classes = [
    GAMER_OT_compute_curvatures,
    GAMER_OT_remove_curvature,
    GAMER_OT_remove_all_curvatures,
    GAMER_OT_plot_curvature,
    GAMER_OT_plot_all_curvatures,
    # GAMER_OT_plot_differences,
    GAMerCurvatureItem,
    GAMerCurvaturesList,
]


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(make_annotations(cls))


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(make_annotations(cls))



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# **************************************************************************

import bpy
from bpy.props import (
    BoolProperty,
    CollectionProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    IntProperty,
    IntVectorProperty,
    PointerProperty,
    StringProperty,
    BoolVectorProperty,
)

import blendgamer.pygamer as g
from blendgamer.util import *
from blendgamer.markers import *


# python imports
import os, sys
import numpy as np
import collections


class GAMER_OT_coarse_dense(bpy.types.Operator):
    bl_idname = "gamer.coarse_dense"
    bl_label = "Coarse Dense"
    bl_description = "Decimate selected dense areas of the mesh"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        try:
            p = context.scene.gamer.surfmesh_improvement_properties
            gmesh = blender_to_gamer(autocorrect_normals=p.autocorrect_normals)
            gmesh.coarse_dense(
                rate=p.dense_rate,
                numiter=p.dense_iter,
                rings=p.rings,
                verbose=p.verbose,
            )
            gamer_to_blender(gmesh)

            self.report({"INFO"}, "GAMer: coarse dense complete")
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class GAMER_OT_coarse_flat(bpy.types.Operator):
    bl_idname = "gamer.coarse_flat"
    bl_label = "Coarse Flat"
    bl_description = "Decimate selected flat areas of the mesh"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        try:

            p = context.scene.gamer.surfmesh_improvement_properties
            gmesh = blender_to_gamer(autocorrect_normals=p.autocorrect_normals)
            gmesh.coarse_flat(
                rate=p.flat_rate, numiter=p.flat_iter, rings=p.rings, verbose=p.verbose
            )
            gamer_to_blender(gmesh)

            self.report({"INFO"}, "GAMer: coarse flat complete")
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class GAMER_OT_smooth(bpy.types.Operator):
    bl_idname = "gamer.smooth"
    bl_label = "Smooth"
    bl_description = "Smooth selected vertices of the mesh"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        try:
            p = context.scene.gamer.surfmesh_improvement_properties
            gmesh = blender_to_gamer(autocorrect_normals=p.autocorrect_normals)
            gmesh.smooth(
                max_iter=p.smooth_iter,
                preserve_ridges=p.preserve_ridges,
                rings=p.rings,
                verbose=p.verbose,
            )
            gamer_to_blender(gmesh)

            self.report({"INFO"}, "GAMer: Smooth Mesh complete")
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class GAMER_OT_normal_smooth(bpy.types.Operator):
    bl_idname = "gamer.normal_smooth"
    bl_label = "Normal Smooth"
    bl_description = "Smooth facet normals of selected faces of the mesh"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        try:
            p = context.scene.gamer.surfmesh_improvement_properties
            gmesh = blender_to_gamer(autocorrect_normals=p.autocorrect_normals)
            gmesh.normalSmooth(p.normSmoothAniso)
            gamer_to_blender(gmesh)

            self.report({"INFO"}, "GAMer: Normal Smooth complete")
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class GAMER_OT_fill_holes(bpy.types.Operator):
    bl_idname = "gamer.fill_holes"
    bl_label = "Fill Holes"
    bl_description = "Triangulate holes"
    bl_options = {"REGISTER", "UNDO"}

    def execute(self, context):
        try:
            p = context.scene.gamer.surfmesh_improvement_properties
            gmesh = blender_to_gamer(autocorrect_normals=p.autocorrect_normals)
            gmesh.fillHoles()
            gamer_to_blender(gmesh)

            self.report({"INFO"}, "GAMer: Fill Holes complete")
            return {"FINISHED"}
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}


class SurfaceMeshImprovementProperties(bpy.types.PropertyGroup):
    dense_rate = FloatProperty(
        name="Thresh",
        default=1,
        min=0.001,
        max=50.0,
        precision=4,
        description="Threshold for coarsening dense areas",
    )

    dense_iter = IntProperty(
        name="Iter",
        default=1,
        min=1,
        max=15,
        description="The number of coarse dense iterations to apply",
    )

    flat_rate = FloatProperty(
        name="Threshold",
        default=0.016,
        min=0.00001,
        max=0.5,
        precision=4,
        description="The rate for coarsening flat areas",
    )

    flat_iter = IntProperty(
        name="Iter",
        default=1,
        min=1,
        max=15,
        description="The number of coarse flat iterations to apply",
    )

    smooth_iter = IntProperty(
        name="Iter",
        default=10,
        min=1,
        max=50,
        description="The number smoothing iterations to apply",
    )

    preserve_ridges = BoolProperty(
        name="Preserve ridges",
        default=False,
        description="Skip flipping of edges which lie on ridges",
    )

    normSmoothAniso = FloatProperty(
        name="Anisotropic control",
        default=1.0,
        min=0,
        description="The degree of anisotropy in LST correction for normal smooth",
    )

    advanced_options = BoolProperty(
        name="Advanced options",
        default=False,
        description="Show additional surface mesh improvement options",
    )

    autocorrect_normals = BoolProperty(
        name="Autocorrect normals",
        default=True,
        description="Automatically flip normals when mesh volume is 'negative'",
    )

    verbose = BoolProperty(
        name="Verbose",
        default=False,
        description="Print additional information to console",
    )

    rings = IntProperty(
        name="LST rings",
        default=2,
        min=1,
        max=5,
        description="The number of neighborhood rings to consider for LST calculation",
    )


classes = [
    GAMER_OT_coarse_dense,
    GAMER_OT_coarse_flat,
    GAMER_OT_smooth,
    GAMER_OT_normal_smooth,
    GAMER_OT_fill_holes,
    SurfaceMeshImprovementProperties,
]


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(make_annotations(cls))


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(make_annotations(cls))



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

"""PyGAMer: Geometry-preserving Adaptive Mesher

PyGAMer is a wrapper around the `GAMer <https://github.com/ctlee/gamer/`__ C++ library.
It enables users to perform mesh generation and manipulation tasks via Python scripts.
"""

import os
import sys
import subprocess
import re

def git_version():
    """Get the version from git describe

    Returns:
        string: version string or None if invalid
    """
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ["SYSTEMROOT", "PATH", "HOME"]:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env["LANGUAGE"] = "C"
        env["LANG"] = "C"
        env["LC_ALL"] = "C"
        out = subprocess.check_output(cmd, stderr=subprocess.STDOUT, env=env)
        return out

    try:
        out = _minimal_ext_cmd(["git", "describe", "--tags", "--dirty", "--always"])
        GIT_REVISION = out.strip().decode("ascii")
    except (subprocess.SubprocessError, OSError):
        GIT_REVISION = None

    return GIT_REVISION

def standardize_version(version_string):
    """Standardize the version string
    
    Args:
        version_string (str): input version string 
    Returns:
        str: standardized version
    """
    VERSION_PATTERN = r"""
        v?
        (?:
            (?:(?P<epoch>[0-9]+)!)?                           # epoch
            (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
            (?P<pre>                                          # pre-release
                [-_\.]?
                (?P<pre_l>(alpha|beta|a|b|c|rc))
                (?P<pre_n>[0-9]+)?
            )?
            (?P<dev>                                          # dev release
                [-_\.]?
                (?P<dev_l>dev)
                [-_\.]?
                (?P<dev_n>[0-9]+)?
            )?
            (?P<meta>
                [-_\.]?
                (?P<commits_since>[0-9]+)?
                [-_\.]?
                (?P<sha>[a-z0-9]*)?
                [-_\.]? 
                (?P<dirty>dirty)?
            )?
        )
    """

    _regex = re.compile(
        r"^\s*" + VERSION_PATTERN + r"\s*$",
        re.VERBOSE | re.IGNORECASE,
    )

    match = _regex.match(version_string)
    if match:
        if match.group("release"):
            version = match.group("release")
            if match.group("pre_l"):
                version += match.group("pre_l")
                if match.group("pre_n"):
                    version += match.group("pre_n")
                else:
                    version += "0"
            if match.group("dev"):
                version += "dev"
                if match.group("dev_n"):
                    version += match.group("dev_n")
                else:
                    version += "0"
    else:
        version = "0.0.0"
    return version

version = git_version()
if version is None:
    # Git describe failed... read version from file
    with open("VERSION", "r") as f:
        version = f.readline()
    version = standardize_version(version)
else:
    version = standardize_version(version)

cmake_args = ["-DBUILD_PYGAMER=ON"]

DOCLINES = __doc__.split("\n")

CLASSIFIERS = """\
Development Status :: 4 - Beta
Environment :: Console
Intended Audience :: Science/Research
License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)
Natural Language :: English
Operating System :: OS Independent
Programming Language :: C++
Programming Language :: Python :: 3 :: Only
Programming Language :: Python :: Implementation :: CPython
Topic :: Scientific/Engineering :: Bio-Informatics
Topic :: Scientific/Engineering :: Chemistry
Topic :: Scientific/Engineering :: Mathematics
Topic :: Scientific/Engineering :: Physics
Topic :: Scientific/Engineering :: Visualization
"""

# If building on readthedocs.io build with unix makefiles
_on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if _on_rtd:
    sys.argv.extend(["-G", "Unix Makefiles"])
    cmake_args.append("-DGAMER_DOCS=ON")

try:
    from skbuild import setup
except ImportError:
    print("\nERROR: scikit-build is required to build from source.", file=sys.stderr)
    print("Please run:", file=sys.stderr)
    print("", file=sys.stderr)
    print("  python -m pip install scikit-build\n", file=sys.stderr)
    print("  -- or --\n", file=sys.stderr)
    print("  conda install scikit-build", file=sys.stderr)
    sys.exit(1)

docs_require = [
    "sphinx",
    "breathe",
    "exhale",
    "sphinx-issues",
    "nbsphinx",
    "jupyter_sphinx",
    "threevis",
]
tests_require = ["pytest"]

setup(
    name="pygamer",
    version=version,
    maintainer="Christopher T. Lee",
    maintainer_email="ctlee@ucsd.edu",
    author="The GAMer Team",
    author_email="ctlee@ucsd.edu",
    url="https://github.com/ctlee/gamer",
    license="LGPLv2+",
    packages=["pygamer"],
    description=DOCLINES[0],
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    platforms=["Windows", "Linux", "Mac OS-X", "Unix"],
    classifiers=[c for c in CLASSIFIERS.split("\n") if c],
    keywords="meshing ",
    cmake_args=cmake_args,
    setup_requires=[
        "setuptools",
        "wheel",
        "scikit-build >= 0.10.0",
        "pytest-runner",
        "cmake >= 3.11",
    ],
    install_requires=["numpy>=1.8.0"],
    extras_require={
        "docs": docs_require,
        "test": tests_require,
    },
    zip_safe=False,
)



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# ***************************************************************************

import array
import mathutils
import math
import bpy
from bpy.props import (
    BoolProperty,
    CollectionProperty,
    EnumProperty,
    FloatProperty,
    FloatVectorProperty,
    IntProperty,
    IntVectorProperty,
    PointerProperty,
    StringProperty,
    BoolVectorProperty,
)
import bmesh

import blendgamer.pygamer as g

import importlib.util

mpl_spec = importlib.util.find_spec("matplotlib")
mpl_found = mpl_spec is not None

if mpl_found:
    from blendgamer.colormap import dataToVertexColor

from blendgamer.colormap_enums import colormap_enums

import blendgamer.report as meshreport
from blendgamer.util import *


## Following ops are from 3D Print Addon
class GAMER_OT_MeshStats_Select_Report(bpy.types.Operator):
    """Select the data associated with this report"""

    bl_idname = "gamer.meshstats_select_report"
    bl_label = "Select Report"
    bl_options = {"INTERNAL"}

    index = IntProperty()

    _type_to_mode = {
        bmesh.types.BMVert: "VERT",
        bmesh.types.BMEdge: "EDGE",
        bmesh.types.BMFace: "FACE",
    }

    _type_to_attr = {
        bmesh.types.BMVert: "verts",
        bmesh.types.BMEdge: "edges",
        bmesh.types.BMFace: "faces",
    }

    def execute(self, context):
        obj = context.edit_object
        info = meshreport.info()
        text, data = info[self.index]
        bm_type, bm_array = data

        bpy.ops.mesh.reveal()
        bpy.ops.mesh.select_all(action="DESELECT")
        bpy.ops.mesh.select_mode(type=self._type_to_mode[bm_type])

        bm = bmesh.from_edit_mesh(obj.data)
        elems = getattr(bm, GAMER_OT_MeshStats_Select_Report._type_to_attr[bm_type])[:]

        try:
            for i in bm_array:
                elems[i].select_set(True)
        except:
            # possible arrays are out of sync
            self.report({"WARNING"}, "Report is out of date, re-run check")
        return {"FINISHED"}


def multiple_obj_warning(self, context):
    if len(context.selected_objects) > 1:
        self.report(
            {"INFO"}, "Multiple selected objects. Only the active one will be evaluated"
        )


# Helper method to get object and run main_check
def execute_check(self, context):
    obj = context.active_object
    info = []
    self.main_check(obj, info)
    meshreport.update(*info)
    multiple_obj_warning(self, context)
    return {"FINISHED"}


class GAMER_OT_MeshStats_Info_Volume(bpy.types.Operator):
    """Report the volume of the active mesh"""

    bl_idname = "gamer.meshstats_info_volume"
    bl_label = "MeshStats Info Volume"

    @staticmethod
    def main_check(obj, info):
        with copiedBMeshContext(obj) as bm:
            volume = 0.0
            for face in bm.faces:
                if len(face.loops) != 3:
                    info.append(
                        ("Cannot compute volume for non triangulated object.", None)
                    )
                    return
                tv0 = face.loops[0].vert.co
                tv1 = face.loops[1].vert.co
                tv2 = face.loops[2].vert.co
                x0 = tv0.x
                y0 = tv0.y
                z0 = tv0.z
                x1 = tv1.x
                y1 = tv1.y
                z1 = tv1.z
                x2 = tv2.x
                y2 = tv2.y
                z2 = tv2.z
                det = (
                    x0 * (y1 * z2 - y2 * z1)
                    + x1 * (y2 * z0 - y0 * z2)
                    + x2 * (y0 * z1 - y1 * z0)
                )
                volume = volume + det
            volume = volume / 6.0
            info.append(("Volume: %s" % clean_float("%.8f" % volume), None))

    def execute(self, context):
        return execute_check(self, context)


class GAMER_OT_MeshStats_Info_Area(bpy.types.Operator):
    """Report the surface area of the active mesh"""

    bl_idname = "gamer.meshstats_info_area"
    bl_label = "MeshStats Info Area"

    @staticmethod
    def main_check(obj, info):
        with copiedBMeshContext(obj) as bm:
            area = sum(f.calc_area() for f in bm.faces)
            info.append(("Area: %s" % clean_float("%.8f" % area), None))

    def execute(self, context):
        return execute_check(self, context)


class GAMER_OT_MeshStats_Check_Solid(bpy.types.Operator):
    """Check for geometry is solid (has valid inside/outside) and correct normals"""

    bl_idname = "gamer.meshstats_check_solid"
    bl_label = "MeshStats Check Solid"
    bl_description = "Check for non-manifolds and inconsistent normals"

    @staticmethod
    def main_check(obj, info):
        with copiedBMeshContext(obj) as bm:
            edges_non_manifold = array.array(
                "i", (i for i, ele in enumerate(bm.edges) if not ele.is_manifold)
            )
            edges_non_contig = array.array(
                "i",
                (
                    i
                    for i, ele in enumerate(bm.edges)
                    if ele.is_manifold and (not ele.is_contiguous)
                ),
            )

            verts_non_manifold = array.array(
                "i", (i for i, ele in enumerate(bm.verts) if not ele.is_manifold)
            )

            info.append(
                (
                    "Non Manifold Edge: %d" % len(edges_non_manifold),
                    (bmesh.types.BMEdge, edges_non_manifold),
                )
            )

            info.append(
                (
                    "Bad Contig. Edges: %d" % len(edges_non_contig),
                    (bmesh.types.BMEdge, edges_non_contig),
                )
            )

            info.append(
                (
                    "Non Manifold Vertices: %d" % len(verts_non_manifold),
                    (bmesh.types.BMVert, verts_non_manifold),
                )
            )

    def execute(self, context):
        return execute_check(self, context)


class GAMER_OT_MeshStats_Check_Intersections(bpy.types.Operator):
    """Check geometry for self intersections"""

    bl_idname = "gamer.meshstats_check_intersect"
    bl_label = "MeshStats Check Intersections"

    @staticmethod
    def main_check(obj, info):
        epsilon = bpy.context.scene.gamer.mesh_quality_properties.intersect_epsilon
        if not obj.data.polygons:
            faces_intersect = array.array("i", ())
        else:
            with copiedBMeshContext(obj) as bm:
                tree = mathutils.bvhtree.BVHTree.FromBMesh(bm, epsilon=epsilon)
                overlap = tree.overlap(tree)
                faces_error = {i for i_pair in overlap for i in i_pair}

                faces_intersect = array.array("i", faces_error)
            info.append(
                (
                    "Intersect Face: %d" % len(faces_intersect),
                    (bmesh.types.BMFace, faces_intersect),
                )
            )

    def execute(self, context):
        return execute_check(self, context)


class GAMER_OT_MeshStats_Check_Degenerate(bpy.types.Operator):
    """Check for degenerate geometry that may not print properly
    (zero area faces, zero length edges)
    """

    bl_idname = "gamer.meshstats_check_degenerate"
    bl_label = "Check Degenerate Faces and Edges"
    bl_description = "Check for zero length/area edges and faces"

    @staticmethod
    def main_check(obj, info):
        threshold = 0
        with copiedBMeshContext(obj) as bm:
            faces_zero = array.array(
                "i",
                (i for i, ele in enumerate(bm.faces) if ele.calc_area() <= threshold),
            )
            edges_zero = array.array(
                "i",
                (i for i, ele in enumerate(bm.edges) if ele.calc_length() <= threshold),
            )

        info.append(
            ("Zero Area Faces: %d" % len(faces_zero), (bmesh.types.BMFace, faces_zero))
        )

        info.append(
            ("Zero Len. Edges: %d" % len(edges_zero), (bmesh.types.BMEdge, edges_zero))
        )

    def execute(self, context):
        return execute_check(self, context)


class GAMER_OT_MeshStats_Check_Wagonwheels(bpy.types.Operator):
    bl_idname = "gamer.meshstats_check_wagonwheels"
    bl_label = "Check for wagon wheels"
    bl_description = "Check for vertices connected to many edges"

    @staticmethod
    def main_check(obj, info):
        n_wagon_edges = bpy.context.scene.gamer.mesh_quality_properties.n_wagon_edges
        with copiedBMeshContext(obj) as bm:
            wagon_edges = array.array(
                "i",
                (
                    i
                    for i, ele in enumerate(bm.verts)
                    if len(ele.link_edges) >= n_wagon_edges
                ),
            )
        info.append(
            (
                "Number of Wagonwheels: %d" % len(wagon_edges),
                (bmesh.types.BMVert, wagon_edges),
            )
        )

    def execute(self, context):
        return execute_check(self, context)


class GAMER_OT_MeshStats_Check_Sharp(bpy.types.Operator):
    bl_idname = "gamer.meshstats_check_sharp"
    bl_label = "Check for small angles"
    bl_description = "Check for faces with small angles"

    @staticmethod
    def main_check(obj, info):
        min_angle = bpy.context.scene.gamer.mesh_quality_properties.min_angle

        with copiedBMeshContext(obj) as bm:
            sharp_list = {
                i
                for i, face in enumerate(bm.faces)
                for loop in face.loops
                if loop.calc_angle() * 180 / math.pi <= min_angle
            }
            sharp = array.array("i", sharp_list)

        info.append(("Sharp faces: %d" % len(sharp), (bmesh.types.BMFace, sharp)))

    def execute(self, context):
        return execute_check(self, context)


class GAMER_OT_MeshStats_Betti_Numbers(bpy.types.Operator):
    bl_idname = "gamer.meshstats_compute_betti"
    bl_label = "Report Betti numbers"
    bl_description = "Compute the first three Betti numbers"

    @staticmethod
    def main_check(obj, info):
        try:
            gmesh = blender_to_gamer(autocorrect_normals=False)
        except Exception as e:
            info.append((str(e), None))
            return
        valid, k, h, v = gmesh.getBettiNumbers()

        info.append(
            (
                "Euler Characteristic: %d"
                % (gmesh.nVertices - gmesh.nEdges + gmesh.nFaces),
                None,
            )
        )

        if valid:
            info.append(("B0 Connected Components: %d" % (k), None))
            info.append(("B1 Holes: %d" % (h), None))
            info.append(("B2 Voids: %d" % (v), None))
        else:
            info.append(("B0 Connected Components: %d" % (k), None))
            info.append(("Higher order Betti numbers undetermined", None))

    def execute(self, context):
        return execute_check(self, context)


class GAMER_OT_MeshStats_Check_All(bpy.types.Operator):
    """Run all checks"""

    bl_idname = "gamer.meshstats_check_all"
    bl_label = "MeshStats Check All"
    bl_description = "Check all"

    def execute(self, context):
        obj = context.active_object
        info = []

        check_classes = (
            GAMER_OT_MeshStats_Info_Volume,
            GAMER_OT_MeshStats_Info_Area,
            GAMER_OT_MeshStats_Check_Wagonwheels,
            GAMER_OT_MeshStats_Check_Sharp,
            GAMER_OT_MeshStats_Check_Solid,
            GAMER_OT_MeshStats_Check_Intersections,
            GAMER_OT_MeshStats_Check_Degenerate,
        )

        try:
            for cls in check_classes:
                cls.main_check(obj, info)
        except Exception as e:
            self.report({"ERROR"}, str(e))
            return {"CANCELLED"}

        meshreport.update(*info)
        multiple_obj_warning(self, context)
        return {"FINISHED"}


class GAMER_OT_write_quality_info(bpy.types.Operator):
    bl_idname = "gamer.write_quality_info"
    bl_label = "Print mesh quality info to files"
    bl_description = "Dump quality info to files"
    bl_options = {"REGISTER"}

    def execute(self, context):
        mqp = bpy.context.scene.gamer.mesh_quality_properties
        for obj in context.selected_objects:
            if obj.type == "MESH":
                fname = mqp.export_path + mqp.export_filebase + "_" + obj.name
                print("Dumping quality info of mesh %s to file %s" % (obj.name, fname))
                try:
                    gmesh = blender_to_gamer(obj=obj)
                    g.printQualityInfo(fname, gmesh)
                except Exception as e:
                    self.report({"ERROR"}, str(e))
                    return {"CANCELLED"}
        return {"FINISHED"}


class MeshQualityReportProperties(bpy.types.PropertyGroup):
    n_wagon_edges = IntProperty(
        name="N Edges",
        default=8,
        min=1,
        description="The number of incident edges to a vertex to be selected",
    )

    export_path = StringProperty(
        name="Export Directory",
        description="Path to directory where files will be created",
        default="//",
        maxlen=1024,
        subtype="DIR_PATH",
    )

    export_filebase = StringProperty(
        name="Filename",
        description="Base name of the files to export",
        default="meshquality",
        maxlen=1024,
        subtype="FILE_NAME",
    )

    min_angle = IntProperty(
        name="Angle Threshold",
        default=15,
        min=0,
        max=180,
        description="Select faces with angles less than this criteria",
    )

    compute_betti = BoolProperty(
        name="Compute Betti Numbers",
        description="Calculate the first 3 betti numbers of a mesh",
        default=False,
    )

    intersect_epsilon = FloatProperty(
        name="Intersection tolerance",
        default=0.00001,
        min=0,
        step=0.000001,
        description="Tolerance use to search for intersecting faces.",
    )

    show_extras = BoolProperty(
        name="Additional Reports",
        default=False,
        description="Show additional report generation options",
    )


classes = [
    GAMER_OT_MeshStats_Select_Report,
    GAMER_OT_MeshStats_Info_Volume,
    GAMER_OT_MeshStats_Info_Area,
    GAMER_OT_MeshStats_Check_Solid,
    GAMER_OT_MeshStats_Check_Intersections,
    GAMER_OT_MeshStats_Check_Degenerate,
    GAMER_OT_MeshStats_Check_Wagonwheels,
    GAMER_OT_MeshStats_Check_Sharp,
    GAMER_OT_MeshStats_Betti_Numbers,
    GAMER_OT_MeshStats_Check_All,
    GAMER_OT_write_quality_info,
    MeshQualityReportProperties,
]


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(make_annotations(cls))


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(make_annotations(cls))



# ***************************************************************************
# This file is part of the GAMer software.
# Copyright (C) 2016-2021
# by Christopher T. Lee and contributors
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http:#www.gnu.org/licenses/>
# or write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA 02111-1307 USA
# ***************************************************************************

import bpy
import bmesh

import blendgamer.report as report
from blendgamer.util import (
    UNSETMARKER,
    make_annotations,
    get_material_by_bnd_id,
    getBndUnsetMat,
)
import blendgamer.pygamer as pygamer


if bpy.app.version < (2, 80, 0):
    REGION = "TOOLS"
    BULB_ICON = "LAMP"
    ADD_ICON = "ZOOMIN"
    REMOVE_ICON = "ZOOMOUT"
    LOVE_ICON = "COLOR_RED"
else:
    REGION = "UI"
    BULB_ICON = "LIGHT"
    ADD_ICON = "ADD"
    REMOVE_ICON = "REMOVE"
    LOVE_ICON = "FUND"


class GAMER_PT_versionerror(bpy.types.Panel):
    bl_label = "GAMer Version Mismatch"
    bl_space_type = "VIEW_3D"
    bl_region_type = REGION
    bl_category = "GAMer"

    @classmethod
    def poll(cls, context):
        return (context.scene is not None) and context.scene.gamer.versionerror != 0

    def draw_header(self, context):
        self.layout.label(text="", icon="CANCEL")

    def draw(self, context):
        layout = self.layout
        layout.alert = True

        if context.scene.gamer.versionerror > 0:
            layout.label(
                text="Warning the current file was generated with a newer version of GAMer!"
            )
            layout.label(
                text="We strongly recommend you update the plugin before manipulating this file."
            )
        elif context.scene.gamer.versionerror < 0:
            layout.label(
                text="Warning the current file was generated with an older version of GAMer could not autoupdate!"
            )
            layout.label(
                text="Please be careful manipulating this file as data may be lost."
            )
            layout.label(text="Use the following converters at your own risk!")
            col = layout.column()
            col.operator("gamer.update_to_2_0_1_from_v_0_1")


class GAMER_PT_surfacemesh(bpy.types.Panel):
    bl_label = "Surface Mesh Conditioning"
    bl_space_type = "VIEW_3D"
    bl_region_type = REGION
    bl_category = "GAMer"
    bl_options = {"DEFAULT_CLOSED"}

    # Panel will be drawn if this returns True
    @classmethod
    def poll(cls, context):
        return context.scene is not None

    def draw_header(self, context):
        self.layout.label(text="", icon="OUTLINER_DATA_MESH")

    def draw(self, context):
        layout = self.layout

        smprops = context.scene.gamer.surfmesh_improvement_properties
        active_obj = context.active_object
        if active_obj and (active_obj.type == "MESH"):
            box = layout.box()
            col = box.column(align=True)
            if not smprops.advanced_options:
                col.prop(smprops, "advanced_options", icon="TRIA_RIGHT", emboss=False)
            else:
                col.prop(smprops, "advanced_options", icon="TRIA_DOWN", emboss=False)
                col.prop(smprops, "autocorrect_normals")
                col.prop(smprops, "verbose")
                col.prop(smprops, "rings")

            col = layout.column()
            col.label(text="Global mesh operations:")
            row = col.row(align=True)
            row.operator("gamer.normal_smooth")
            row.prop(smprops, "normSmoothAniso")
            col.operator("gamer.fill_holes")

            if active_obj.mode == "EDIT":
                col = layout.column()
                col.label(text="Local operations (applied to selection only): ")

                col = layout.column(align=True)
                col.operator("gamer.coarse_dense")
                rowsub = col.row(align=True)
                rowsub.prop(smprops, "dense_rate")
                rowsub.prop(smprops, "dense_iter")

                col = layout.column(align=True)
                col.operator("gamer.coarse_flat")
                rowsub = col.row(align=True)
                rowsub.prop(smprops, "flat_rate")
                rowsub.prop(smprops, "flat_iter")

                col = layout.column(align=True)
                col.operator("gamer.smooth")
                rowsub = col.row(align=True)
                rowsub.prop(smprops, "smooth_iter")
                rowsub.prop(smprops, "preserve_ridges", expand=True)
            else:
                col = layout.column()
                col.label(
                    text="Change to Edit Mode to enable local improvement options",
                    icon="INFO",
                )
        else:
            layout.label(
                text="Select a mesh object to use GAMer mesh processing features",
                icon="HAND",
            )


# class GAMER_PT_advanced_options(bpy.types.Panel):
#     bl_label = "Advanced Options"
#     bl_parent_id = 'GAMER_PT_surfacemesh'
#     bl_space_type = 'VIEW_3D'
#     bl_region_type = REGION
#     bl_category = "GAMer"
#     bl_options = {'DEFAULT_CLOSED'}


#     def draw(self, context):
#         layout = self.layout

#         smprops = context.scene.gamer.surfmesh_improvement_properties
#         active_obj = context.active_object

#         col = layout.column(align=True)
#         col.prop(smprops, "autocorrect_normals")
#         col.prop(smprops, "verbose")
#         col.prop(smprops, "rings")


class GAMER_PT_mesh_quality(bpy.types.Panel):
    bl_label = "Mesh Quality Reporting"
    bl_space_type = "VIEW_3D"
    bl_region_type = REGION
    bl_category = "GAMer"
    bl_options = {"DEFAULT_CLOSED"}

    _type_to_icon = {
        bmesh.types.BMVert: "VERTEXSEL",
        bmesh.types.BMEdge: "EDGESEL",
        bmesh.types.BMFace: "FACESEL",
    }

    # Panel will be drawn if this returns True
    @classmethod
    def poll(cls, context):
        return context.scene is not None

    # This method is from 3D Print Addon
    @staticmethod
    def draw_report(layout, context):
        """Display Reports"""
        info = report.info()
        if info:
            obj = context.edit_object

            layout.label(text="Mesh Stats Report:")
            box = layout.box()
            col = box.column(align=False)
            # box.alert = True
            for i, (text, data) in enumerate(info):
                if obj and data and data[1]:
                    bm_type, _ = data
                    col.operator(
                        "gamer.meshstats_select_report",
                        text=text,
                        icon=GAMER_PT_mesh_quality._type_to_icon[bm_type],
                    ).index = i
                else:
                    col.label(text=text)

    def draw(self, context):
        layout = self.layout
        # Mesh quality properties object
        qProps = context.scene.gamer.mesh_quality_properties

        col = layout.column(align=True)
        col.operator("gamer.meshstats_check_solid")
        col.operator("gamer.meshstats_check_degenerate")
        col.operator("gamer.meshstats_check_intersect")
        col.operator("gamer.meshstats_compute_betti")

        row = col.row(align=True)
        row.operator("gamer.meshstats_check_wagonwheels")
        row.prop(qProps, "n_wagon_edges")

        row = col.row(align=True)
        row.operator("gamer.meshstats_check_sharp")
        row.prop(qProps, "min_angle")
        col = layout.column()
        col.label(text="Mesh analysis:")
        col.operator("gamer.meshstats_check_all", text="Generate Mesh Report")

        GAMER_PT_mesh_quality.draw_report(layout, context)

        col = layout.column()
        col.label(text="Curvature Estimation:")

        if context.scene.gamer.matplotlib_found:
            active_obj = context.active_object
            if active_obj and (active_obj.type == "MESH"):
                curveProp = active_obj.gamer.curvatures
                row = col.row(align=True)
                row.operator("gamer.compute_curvatures")
                row.prop(curveProp, "algorithm", text="")

                row = layout.row()
                row.label(text="Computed Curvatures:", icon="FACESEL")
                row = layout.row()
                col = row.column()
                col.template_list(
                    "GAMER_UL_curvature_list",
                    "GAMer Curvature List",
                    curveProp,
                    "curvature_list",
                    curveProp,
                    "active_index",
                    rows=2,
                    type="DEFAULT",
                )
                col = row.column(align=True)
                col.operator("gamer.remove_curvature", icon=REMOVE_ICON, text="")
                col.operator("gamer.remove_all_curvatures", icon="X", text="")

                crv = curveProp.get_active_index()
                if crv != None:
                    col = layout.column()
                    row = col.row()
                    row.label(text="Set active curvature properties:")

                    row = col.row()
                    row.prop(crv, "colormap")
                    row.prop(crv, "limitsArePercentiles")

                    col = layout.column(align=True)
                    row = col.row(align=True)
                    row.prop(crv, "mixpoint")
                    row.prop(crv, "curveIter")
                    row = col.row(align=True)
                    row.prop(crv, "minCurve")
                    row.prop(crv, "maxCurve")

                    row = layout.row()
                    row.label(text="Plot Settings:")
                    row = layout.row()
                    row.prop(curveProp, "showplots")
                    row.prop(curveProp, "saveplots")

                    row = layout.row()
                    row.operator("gamer.plot_curvature")
                    row.operator("gamer.plot_all_curvatures")
                    # row = layout.row()
                    # row.operator("gamer.plot_differences")
            else:
                col.label(
                    text="Select a mesh object to enable estimation of curvatures",
                    icon=BULB_ICON,
                )
        else:
            col.label(text="Curvature estimations require matplotlib.", icon=BULB_ICON)

        box = layout.box()
        col = box.column()

        if not qProps.show_extras:
            col.prop(qProps, "show_extras", icon="TRIA_RIGHT", emboss=False)
        else:
            col.prop(qProps, "show_extras", icon="TRIA_DOWN", emboss=False)
            col.prop(qProps, "export_path")
            col.prop(qProps, "export_filebase")
            col.operator("gamer.write_quality_info")


class GAMER_UL_curvature_list(bpy.types.UIList):
    def draw_item(
        self, context, layout, data, item, icon, active_data, active_propname, index
    ):
        """
        Draw the UI list for boundary markers
        """
        row = layout.row()
        row.label(text=layout.enum_item_description(item, "algorithm", item.algorithm))
        # row.prop_enum(item, 'algorithm', item.algorithm, text=layout.enum_item_description(item, 'algorithm', item.algorithm))
        row.prop_enum(item, "curvatureType", item.curvatureType)


class GAMER_PT_boundary_marking(bpy.types.Panel):
    bl_label = "Boundary Marking"
    bl_space_type = "VIEW_3D"
    bl_region_type = REGION
    bl_category = "GAMer"
    bl_options = {"DEFAULT_CLOSED"}

    @classmethod
    def poll(cls, context):
        return context.scene is not None

    def draw_header(self, context):
        self.layout.label(text="", icon="TPAINT_HLT")

    def draw(self, context):
        layout = self.layout

        active_obj = context.active_object
        if active_obj and (active_obj.type == "MESH"):
            markerProp = active_obj.gamer.markers

            row = layout.row()
            row.label(text="Unmarked marker value = %d" % (UNSETMARKER))
            row = layout.row()

            bnd_unset_mat = getBndUnsetMat()
            row.prop(bnd_unset_mat, "diffuse_color", text="Unmarked boundary color")

            row = layout.row()
            row.label(text="Defined Boundaries:", icon="FACESEL")
            row = layout.row()
            col = row.column()
            col.template_list(
                "GAMER_UL_boundary_list",
                "GAMer Boundary List",
                markerProp,
                "boundary_list",
                markerProp,
                "active_bnd_index",
                rows=2,
                type="DEFAULT",
            )
            col = row.column(align=True)
            col.operator("gamer.add_boundary", icon=ADD_ICON, text="")
            col.operator("gamer.remove_boundary", icon=REMOVE_ICON, text="")
            col.operator("gamer.remove_all_boundaries", icon="X", text="")

            # Could have boundary item draw itself in new row here:
            active_bnd = markerProp.get_active_boundary()
            if active_bnd:
                row = layout.row()
                row.label(text="Set active boundary properties:")

                # Row to update active_bnd name
                row = layout.row()
                row.prop(active_bnd, "boundary_name")
                # Row to update marker value
                row = layout.row()
                row.label(text="Marker:")
                row.prop(active_bnd, "marker", text="")  # suppress default txt

                row = layout.row()
                if active_obj.mode == "OBJECT":
                    row.label(
                        text="Change to Edit Mode to enable boundary assignment",
                        icon="INFO",
                    )

            if active_obj.mode == "EDIT" and active_bnd:
                row = layout.row()
                sub = row.row(align=True)
                sub.operator("gamer.assign_boundary_faces", text="Assign")
                sub.operator("gamer.remove_boundary_faces", text="Remove")

                sub = row.row(align=True)
                sub.operator("gamer.select_boundary_faces", text="Select")
                sub.operator("gamer.deselect_boundary_faces", text="Deselect")

                layout.separator()
                row = layout.row()
                row.operator(
                    "gamer.select_all_boundary_faces",
                    text="Select All Marked Boundaries",
                )
                row.operator(
                    "gamer.deselect_all_boundary_faces",
                    text="Deselect All Marked Boundaries",
                )
        else:
            layout.label(
                text="Select a mesh object to use GAMer boundary marking features",
                icon="HAND",
            )


class GAMER_UL_boundary_list(bpy.types.UIList):
    def draw_item(
        self, context, layout, data, item, icon, active_data, active_propname, index
    ):
        """
        Draw the UI list for boundary markers
        """
        if item.status:
            layout.label(text=item.boundary_name, icon="ERROR")
        else:
            layout.label(text=item.boundary_name)

        # Show the color swatch in last section only
        if bpy.app.version < (2, 80, 0):
            split = layout.split(percentage=0.5, align=True)
        else:
            split = layout.split(factor=0.5, align=True)
        col = split.column()
        col = split.column()

        bnd_mat = get_material_by_bnd_id(item.boundary_id)
        col.prop(bnd_mat, "diffuse_color", text="")
        # Currently no check for if the referenced material does not exist
        # Also doesn't check if it's associated in a material slot for the
        # active object.


class GAMER_UL_domain(bpy.types.UIList):
    def draw_item(
        self, context, layout, data, item, icon, active_data, active_propname, index
    ):
        """The draw_item function is called for each item of the collection that is visible in the list.

        Args:
            context ([type]): the current Blender context
            layout ([type]): the local layout
            data ([type]): the RNA object containing the collection
            item ([type]): the current drawn item of the collection
            icon ([type]): the "computed" icon for the item (as an integer, because some objects like materials or textures  have custom icons ID, which are not available as enum items).
            active_data ([type]): ct containing the active property for the collection (i.e. integer pointing to the active item of the collection).
            active_propname ([type]): the name of the active property (use 'getattr(active_data, active_propname)')
            index ([type]): index of the current item in the collection
        """
        # The item will be a GAMerTetDomainPropertyGroup
        # Let it draw itself in a new row:
        item.draw_item_in_row(layout.row())


class GAMER_PT_tetrahedralization(bpy.types.Panel):
    bl_label = "Tetrahedralization"
    bl_space_type = "VIEW_3D"
    bl_region_type = REGION
    bl_category = "GAMer"
    bl_options = {"DEFAULT_CLOSED"}

    @classmethod
    def poll(cls, context):
        return context.scene is not None

    def draw_header(self, context):
        self.layout.label(text="", icon="MESH_ICOSPHERE")

    def draw(self, context):
        layout = self.layout

        tetprops = context.scene.gamer.tet_group

        row = layout.row()
        row.label(text="Domains")

        row = layout.row()
        col = row.column()

        col.template_list(
            "GAMER_UL_domain",
            "",
            tetprops,
            "domain_list",
            tetprops,
            "active_domain_index",
            rows=2,
        )

        col = row.column(align=True)
        col.operator("gamer.tet_domain_add", icon=ADD_ICON, text="")
        col.operator("gamer.tet_domain_remove", icon=REMOVE_ICON, text="")
        col.operator("gamer.tet_domain_remove_all", icon="X", text="")
        row = col.row()
        row.operator("gamer.cleanup_domains", icon="GHOST_DISABLED", text="")

        if len(tetprops.domain_list) > 0:
            domain = tetprops.domain_list[tetprops.active_domain_index]

            # row = layout.row()
            # row.label ( "Active Index = " + str ( self.active_domain_index ) + ", ID = " + str ( domain.domain_id ) )

            domain.draw_layout(layout)

            box = layout.box()
            row = box.row(align=True)
            row.alignment = "LEFT"
            if not tetprops.show_settings:
                row.prop(tetprops, "show_settings", icon="TRIA_RIGHT", emboss=False)
            else:
                row.prop(tetprops, "show_settings", icon="TRIA_DOWN", emboss=False)

                row = box.row()
                row.prop(tetprops, "export_path")
                row = box.row()
                row.prop(tetprops, "export_filebase")

                row = box.row()
                col = row.column()
                col.prop(tetprops, "min_dihedral")
                col = row.column()
                col.prop(tetprops, "max_aspect_ratio")

                # row = box.row()
                # row.prop ( self, "ho_mesh" )

                row = box.row()
                row.label(text="Output Formats:")

                row = box.row()
                sbox = row.box()

                row = sbox.row()
                col = row.column()
                col.prop(tetprops, "dolfin")
                col = row.column()
                col.prop(tetprops, "paraview")
                col = row.column()
                col.prop(tetprops, "comsol")

            row = layout.row()
            icon = "PROP_OFF"
            if tetprops.dolfin or tetprops.paraview:
                icon = "PROP_ON"
            row.operator("gamer.tetrahedralize", text="Tetrahedralize", icon=icon)
            if len(tetprops.status) > 0:
                row = layout.row()
                row.label(text=tetprops.status, icon="ERROR")


class GAMER_PT_version(bpy.types.Panel):
    bl_label = "BlendGAMer Info"
    bl_space_type = "VIEW_3D"
    bl_region_type = REGION
    bl_category = "GAMer"
    bl_options = {"DEFAULT_CLOSED"}

    @classmethod
    def poll(cls, context):
        return context.scene is not None

    def draw_header(self, context):
        self.layout.label(text="", icon=LOVE_ICON)

    def draw(self, context):
        layout = self.layout
        layout.alert = True

        layout.label(text="BlendGAMer %s" % (pygamer.__version__()))
        layout.operator(
            "wm.url_open", text="How to Acknowledge"
        ).url = "https://gamer.readthedocs.io/en/latest/#acknowleding-the-use-of-gamer-in-your-work"


classes = [
    GAMER_PT_versionerror,
    GAMER_PT_surfacemesh,
    # GAMER_PT_advanced_options,
    GAMER_PT_mesh_quality,
    GAMER_UL_curvature_list,
    GAMER_PT_boundary_marking,
    GAMER_UL_boundary_list,
    GAMER_UL_domain,
    GAMER_PT_tetrahedralization,
    GAMER_PT_version,
]


def register():
    from bpy.utils import register_class

    for cls in classes:
        register_class(make_annotations(cls))


def unregister():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(make_annotations(cls))
